import re
from dataEngine import Common, Logger, Driver


logger = Logger.get()
ChannelsList = {}
ChannelsListByName = {}
ua_regex = re.compile(r'[^a-zA-Z0-9+\-_. ]')


class ChanType:
    SIP = 1
    LOCAL = 2


CHAN_INFO_FIELDS = ("userId", "email", "userExtension", "groupId", "groupName", "phone", "name", "company",
                    "userDepartment", "device")


class Record(Common.PbxDataObject):
    __slots__ = ("fileName", "isSystem", "args", "onPause")

    def __init__(self, fileName, isSystem, args):
        self.fileName = fileName
        self.isSystem = isSystem
        self.args = args
        self.onPause = False

    def is_system(self):
        return self.isSystem

    def is_on_pause(self):
        return self.onPause

    def get_file_name(self):
        return self.fileName

    def get_args(self):
        # as "Action: Record" in the current implementation is sent after 'START_BRIDGE' event only we mustn't
        # send 'b' option to allow Callweaver to start a recording immediately
        args = self.args.replace("b", "")
        return args


class Channel(Common.PbxDataObject):
    __slots__ = ("uid", "lid", "userId", "email", "userExtension", "groupId", "groupName", "phone", "name", "company",
                 "chanType", "chanName", "context", "removed", "userDepartment", "device", "record", "splitRecordDst")

    def __init__(self, uid, lid, userID, email, userExtension, groupID, groupName, phone, name, chanType, chanName,
                 context, userDepartment):
        self.uid = uid
        self.lid = lid
        self.userId = userID
        self.email = email
        self.userExtension = userExtension
        self.groupId = groupID
        self.groupName = groupName
        self.phone = phone
        self.name = name
        self.company = None
        self.chanType = chanType
        self.chanName = chanName
        self.context = context
        self.removed = False
        self.userDepartment = userDepartment
        self.device = None
        self.record = None
        self.splitRecordDst = None

    def __str__(self):
        return "%s(%s) - %s (%s)" % (self.uid, self.lid, self.chanName, self.get_info())

    def get_info(self):
        return {attr: self.get(attr) for attr in CHAN_INFO_FIELDS if self.get(attr) is not None}

    def set_info(self, info):
        for key, value in info.items():
            self.set(key, value)

    def set_record(self, fileName=None, isSystemRecord=None, sysRecordArgs=""):
        if self.record is None:
            self.record = Record(fileName, isSystemRecord, sysRecordArgs)
        else:
            self.record.set("fileName", fileName)
            self.record.set("isSystem", isSystemRecord)

    def set_split_record(self, dest=""):
        if self.record is None:
            logger.debug("Skip setting splitRecordDst as this channel is not being recorded.")
            return

        if self.splitRecordDst is not None:
            logger.debug("Skip setting splitRecordDst as it has been already set.")
            return

        logger.debug("Set splitRecordDst for the following channel: %s" % self)
        self.splitRecordDst = dest
        return

    def split_record(self):
        record = self.record
        if record is None:
            return False

        if self.splitRecordDst is None:
            logger.debug("Skip splitting a recording as splitRecordDst is not set." % self)
            return False

        logger.info("In the conversation '%s' we have %s record for channel %s (dest: %s) and we have to split "
                    "the conversation, let's stop current record and start a new one"
                    % (self.lid, "system" if record.is_system() else "personal", self.chanName, self.splitRecordDst))
        Driver.send_action(
            "Record",
            Channel=self.chanName,
            Type="split",
            Status="yes",
            Args="{},{},{},{}".format("pbx" if record.is_system() else "personal",
                                      "pause" if record.is_on_pause() else "",
                                      self.splitRecordDst, record.get_args())
        )
        return True

    def set_record_pause_state(self, state):
        record = self.record
        if record is None:
            return
        record.set("onPause", state)

    def set(self, param, value):
        if param == "device" and value:
            value = ua_regex.sub('', value)
        super().set(param, value)

    def isRemoved(self):
        return self.removed

    def __eq__(self, other):
        return all(other and self.get(p) == other.get(p) for p in CHAN_INFO_FIELDS)


def add(chan, pbx="callweaver"):
    if pbx not in ChannelsList:
        ChannelsList[pbx] = {}
        ChannelsListByName[pbx] = {}
    ChannelsList[pbx][chan.uid] = chan
    ChannelsListByName[pbx][chan.chanName] = chan


def get(uid, pbx="callweaver"):
    if pbx in ChannelsList:
        return ChannelsList[pbx].get(uid)


def get_by_name(chanName, pbx="callweaver"):
    if pbx in ChannelsList:
        return ChannelsListByName[pbx].get(chanName)


def uid_masquerade(uid1, uid2, lid, pbx="callweaver"):
    chan1 = ChannelsList[pbx].get(uid1)
    chan2 = ChannelsList[pbx].get(uid2)
    if chan1 is None or chan2 is None:
        return
    chan1.set("uid", uid2)
    chan1.set("lid", lid)
    chan2.set("uid", uid1)
    chan2.set("lid", lid)
    ChannelsList[pbx][uid1] = chan2
    ChannelsList[pbx][uid2] = chan1
    ChannelsListByName[pbx][chan1.chanName] = chan2
    ChannelsListByName[pbx][chan2.chanName] = chan1


def end(uid, pbx="callweaver"):
    if pbx in ChannelsList:
        chan = ChannelsList[pbx].pop(uid, None)
        if chan:
            ChannelsListByName[pbx].pop(chan.chanName, None)


def print_current_state():
    logger.notice(get_current_state())


def end_all():
    for pbx in list(ChannelsList):
        if pbx in ChannelsList:
            for chan in list(ChannelsList[pbx].keys()):
                if chan in ChannelsList[pbx].keys():
                    end(chan, pbx)


def get_current_state():
    lines = ["All channels details:\r\n"]
    for pbx in ChannelsList:
        lines.append("%s PBX: %s" % ("=" * 10, pbx))
        for channel in ChannelsList[pbx]:
            lines.append("----- " + str(ChannelsList[pbx][channel]))
    lines.append("\r\n")
    return "\r\n".join(lines)
