import time
from dataEngine import Common, Event, Channel, Logger

logger = Logger.get()
ConversationsList = {}
ChannelsFromWmsNetChannels = {}
ChannelsToWmsNetChannels = {}


class Conversation(Common.PbxDataObject):
    __slots__ = ("ID", "status", "caller", "callee", "app", "destination", "service", "trunk", "queue", "page", "tags", "flags", "splitReason",
                 "transferType", "mergeWith", "cause", "causeStr", "who", "lookupInfo", "last_event", "event", "wmsnet_channels_app",
                 "skip_events_for_wmsnet_calls", "skip_moh_event", "conversation_bridged")

    def __init__(self, ID, caller, destination, company=None):
        self.ID = ID
        self.status = "connecting"
        self.caller = caller
        self.callee = None
        self.app = None
        self.destination = destination
        self.service = None
        self.trunk = None
        self.queue = None
        self.page = None
        self.tags = None
        self.flags = []
        self.splitReason = None
        self.transferType = None
        self.mergeWith = None
        self.cause = None
        self.causeStr = None
        self.who = None
        self.lookupInfo = None
        self.last_event = None
        self.wmsnet_channels_app = None
        self.event = Event.Event(ID, self.status, caller, destination, company)
        self.skip_events_for_wmsnet_calls = False
        self.skip_moh_event = False
        self.conversation_bridged = False

    def __eq__(self, other):
        return self.ID == other.ID

    def __str__(self):
        return "Conversation object (lid %s). Caller: %s; Callee: %s; app: %s, wmsnet_channels_app: %s" % (self.ID, self.caller, self.callee, self.app, self.wmsnet_channels_app if self.wmsnet_channels_app else "None")

    def setWmsNet(self, chan, remoteApp, isFromWmsNet):
        if not self.wmsnet_channels_app:
            self.wmsnet_channels_app = {}
        self.wmsnet_channels_app[chan] = {"app": remoteApp, "isFromWmsNet": isFromWmsNet}

    def get_wmsnet_app(self, chan):
        return self.wmsnet_channels_app[chan] if self.wmsnet_channels_app and chan in self.wmsnet_channels_app else None

    def set(self, param, value):
        if not value or (param == "destination" and value == "s"):
            return
        current_value = self.get(param)
        super().set(param, value)
        if value == current_value:
            return
        self.event.set(param, value)
        if self.event.isChanged() and param in ("status", "callee", "caller"):
            self.conv_update()

    def mark_remove_participant(self, param):
        if param == "callee" and self.callee:
            self.callee.set("removed", True)
        if param == "caller" and self.caller:
            self.caller.set("removed", True)

    def link_channel(self, participant, chan):
        if participant not in ("caller", "callee"):
            return
        # merge callee info looked up before callee channel has started
        if self.lookupInfo and participant == "callee":
            chan.set_info(self.lookupInfo)
            self.lookupInfo = None
        self.set(participant, chan)

    def set_participant_info(self, participant, info):
        if participant not in ("caller", "callee"):
            return
        chan = self.get(participant)

        if "device" in info:  # processing SIPUserAgent event
            chan.set("device", info["device"])
            self.event.set(participant, chan)

            if self.status == "talking":  # this is a hack for a semi-attended transfer done through Collaboration
                logger.debug("Force update of caller/callee as conversation status is already 'talking'.")
                self.conv_update()

            if self.transferType == "attended":  # this is a hack for a semi-attended transfer done through WP
                conv_obj = get(self.get("mergeWith"))
                if conv_obj:
                    role = conv_obj.get_participant_role(chan.get("uid"))
                    if role:
                        conv_obj.event.set(role, chan)
                        logger.debug("Found another conversation with the same channel, updating...")
                        if conv_obj.status == "talking":
                            # this is for a rare case when attended transfer is switched to a semi-attended too fast
                            logger.info("Forcing an update of a '%s' due to user-agent detection logic." % role)
                            conv_obj.conv_update()

            return

        if len(info["phone"]) > 8 and info["phone"].startswith("00") and info["phone"][2] != "0":
            info["phone"] = "+%s" % info["phone"][2:]

        if participant == "callee":
            # store callee info, it will be processed upon callee channel start
            if len(info.keys()) == 1:  # processing NumberNormalization event
                if self.lookupInfo:    # update just a number to a normalized one in case of lookup was done in dialplan
                    self.lookupInfo["phone"] = info["phone"]
                else:  # copy channel data for future usage in case of a cascade of trunks when a first dial failed
                    self.lookupInfo = dict()
                    self.lookupInfo["phone"] = info["phone"]
                    self.lookupInfo["name"] = chan.name
                    self.lookupInfo["company"] = chan.company
                    self.lookupInfo["email"] = chan.email
            else:                      # processing LookUp event
                self.lookupInfo = info
        else:
            if (info.get("company") and info["company"] != chan.company) or \
                    (info.get("email") and info["email"] != chan.email) or \
                    (info.get("name") and info["name"] != chan.name) or \
                    (info.get("phone") and info["phone"] != chan.phone):
                chan.set_info(info)
                self.event.set(participant, chan)
                self.conv_update()

    def set_flags(self, value):
        self.flags.append(value)
        self.event.flags.append(value)

    def set_app(self, app, destination):
        if app not in ("Dial", "Queue", "Page", "ConfBridge", "VoiceMail", "CloudConference"):
            return

        if app == "Queue":
            self.skip_moh_event = True

        current_app = self.app
        current_dest = self.destination
        self.app = app
        self.event.app = app
        if app != "Queue":
            self.set("destination", destination)
        if app == "VoiceMail":
            self.set_flags("voicemail")
        if app == "CloudConference":
            self.set_flags("wizyConference")
        if app == "ConfBridge":
            self.set_flags("pbxConference")
        if (self.last_event == "split" or not current_app) and (current_dest != destination or app == "VoiceMail"):
            self.conv_update()
        elif (current_app != app and app in ("Queue", "Page", "VoiceMail")) or current_dest != destination:
            self.conv_split("dialplan", destination)

    def conv_update(self):
        self.last_event = "update"
        self.event.push("CALL.UPDATE")

    def conv_split(self, reason, dest=""):
        if self.caller:
            self.caller.set_split_record(dest)
        if self.callee:
            self.callee.set_split_record(dest)

        self.last_event = "split"
        self.status = "connecting"
        self.event.status = "connecting"
        self.splitReason = reason
        self.event.reason = reason
        self.queue = None
        self.event.push("CALL.SPLIT")
        self.skip_events_for_wmsnet_calls = False
        self.conversation_bridged = False
        self.skip_moh_event = False

    def record_start(self, chanObj=None, fileName=None, isSystemRecord=False):
        if not chanObj:
            logger.error("No chanObj for record start event")
            return
        if not fileName:
            logger.error("No file name in record start event")
            return
        chanObj.set_record(fileName=fileName, isSystemRecord=isSystemRecord)
        self.event.record_file_name = fileName
        self.event.push("CALL.RECORD_START")

    def record_stop(self, chanObj=None, fileName=None):
        if chanObj is None:
            logger.info("We don't have channel to send call record stop event")
            return
        if chanObj.record is None:
            return
        if not fileName:
            fileName = chanObj.record.get_file_name()
        self.event.record_file_name = fileName
        self.event.push("CALL.RECORD_STOP")
        chanObj.record = None

    def call_live(self):
        self["event"].push("CALL.LIVE")

    def get_participant_role(self, UniqueID):
        res = None
        callerObj = self.get("caller")
        if callerObj and callerObj.get("uid") == UniqueID:
            res = "caller"
        calleeObj = self.get("callee")
        if calleeObj and calleeObj.get("uid") == UniqueID:
            res = "callee"
        return res

    def getWmsNet(self, chan):
        if chan is None:
            return None
        elif isinstance(chan, Channel.Channel):
            chan = chan.chanName
        return self.get_wmsnet_app(chan)


def add_wmsnet_channels(chan, app, is_from_wms_net):
    if is_from_wms_net:
        ChannelsFromWmsNetChannels[chan] = app.capitalize() if isinstance(app, str) else True
    else:
        ChannelsToWmsNetChannels[chan] = app.capitalize() if isinstance(app, str) else True


def pop_wmsnet_channels(chan):
    if chan in ChannelsFromWmsNetChannels:
        return ChannelsFromWmsNetChannels.pop(chan), True
    if chan in ChannelsToWmsNetChannels:
        return ChannelsToWmsNetChannels.pop(chan), False
    return None, None


def add(conv, pbx="callweaver"):
    if pbx not in ConversationsList:
        ConversationsList[pbx] = {}
    ConversationsList[pbx][conv.ID] = conv
    conv.event.push("CALL.START")


def get(ID, pbx="callweaver"):
    if pbx in ConversationsList:
        return ConversationsList[pbx].get(ID)


def get_by_chan(uid, pbx):
    if pbx not in ConversationsList:
        return

    for conv_id in ConversationsList[pbx]:  # search algorithm might be improved
        conv_obj = ConversationsList[pbx][conv_id]
        if (conv_obj.callee and conv_obj.callee.get("uid") == uid) or \
                (conv_obj.caller and conv_obj.caller.get("uid") == uid):
            return conv_obj


def end(ID, pbx="callweaver"):
    conv = ConversationsList[pbx].pop(ID, None)
    if conv is None:
        return

    for role in ("caller", "callee"):
        if conv.get(role):
            conv.record_stop(conv.get(role))

    conv.event.push("CALL.END")
    conv.status = "end"


def print_current_state():
    logger.notice(get_current_state())


def update_active_calls():
    while True:
        time.sleep(30)
        for pbx in list(ConversationsList.keys()):
            for conv in list(ConversationsList.get(pbx, {}).keys()):
                conv_obj = get(conv, pbx)
                try:
                    if conv_obj and Event.get_time() - conv_obj["event"]["time"] > 300:
                        conv_obj.call_live()
                except Exception as e:
                    logger.error("Something goes wrong for the sending CALL.LIVE event (conv: %s, pbx: %s, Exception: %s)" % (conv, pbx, e))


def end_all():
    for pbx in list(ConversationsList):
        if pbx in ConversationsList:
            for ID in list(ConversationsList[pbx].keys()):
                if ID in ConversationsList[pbx].keys():
                    conv = ConversationsList[pbx].get(ID)
                    conv.set("cause", "-1")
                    conv.set("causeStr", "PBX engine has been restarted")
                    conv.set("who", "system")
                    end(ID, True, pbx)


def get_current_state():
    lines = ["All conversations details:\r\n"]
    for pbx in ConversationsList:
        lines.append("%s PBX: %s" % ("=" * 10, pbx))
        for channel in ConversationsList[pbx]:
            lines.append("----- " + str(ConversationsList[pbx][channel]))
    lines.append("\r\n")
    return "\r\n".join(lines)
