import json
import time
from dataEngine import Common, Aws, AppConfig, Logger
from .Driver import ConnectorType

logger = Logger.get()
time_from_pcap = 0


def get_time():
    if AppConfig.get("rewrite_traces"):
        return time_from_pcap
    else:
        return time.time()


def set_time(t):
    if t:
        global time_from_pcap
        time_from_pcap = t


class Event(Common.PbxDataObject):
    __slots__ = ("ID", "name", "status", "service", "app", "destination", "trunk", "queue", "page", "tags", "reason",
                 "transferType", "record_file_name", "mergeWith", "cause", "causeStr", "who", "time", "_caller", "_callee",
                 "_flags", "_changed", "_company")

    def __init__(self, ID, status, caller, destination, company=None):
        self.ID = ID
        self.name = None
        self.status = status
        self.service = None
        self.app = None
        self.destination = destination
        self.trunk = None
        self.queue = None
        self.page = None
        self.tags = None
        self.reason = None
        self.transferType = None
        self.record_file_name = None
        self.mergeWith = None
        self.cause = None
        self.causeStr = None
        self.who = None
        self.time = get_time()
        self._caller = caller
        self._callee = {}
        self._flags = []
        self._changed = False
        self._company = company

    def set(self, param, value):
        self._changed = (value != getattr(self, param))
        if value:
            setattr(self, param, value)

    @property
    def caller(self):
        return self._caller.get_info() if self._caller else {}

    @caller.setter
    def caller(self, new_val):
        self._caller = new_val

    @property
    def callee(self):
        return self._callee.get_info() if self._callee else {}

    @callee.setter
    def callee(self, new_val):
        self._callee = new_val

    @property
    def flags(self):
        return self._flags

    @flags.setter
    def flags(self, new_val):
        self._changed = (new_val != self._flags)
        if new_val is None:
            self._flags = []
        else:
            self._flags = new_val

    def isChanged(self):
        return self._changed

    def _get_data(self, attrs):
        rv = {}
        for attr in attrs:
            value = self.get(attr)
            if value:
                rv[attr] = value
                setattr(self, attr, None)
        return rv

    # START/UPDATE
    def get_data(self):
        attrs = ["status", "caller", "callee", "service", "destination", "trunk", "queue", "tags", "flags", "page"]
        return self._get_data(attrs)

    def get_data_split(self):
        attrs = ["reason", "service", "transferType", "mergeWith", "status", "caller", "callee", "destination",
                 "trunk", "queue", "tags", "flags"]
        return self._get_data(attrs)

    def get_data_agent(self):
        return {}

    def get_data_record(self):
        rv = {}
        if self.record_file_name is not None:
            rv["fileName"] = self.record_file_name
            self.record_file_name = None
        return rv

    def get_data_end(self):
        attrs = ["cause", "causeStr", "who"]
        data = self._get_data(attrs)
        if data.get("cause") == "0":
            data.pop("cause")
            data.pop("causeStr")
        return data

    def push(self, name):
        # print("\n\n>>> Event.push, self: {}\n\n".format(self))
        data = None
        self.time = get_time()
        if name in ("CALL.START", "CALL.UPDATE"):
            data = self.get_data()
        elif name == "CALL.SPLIT":
            data = self.get_data_split()
        elif name in ("CALL.RECORD_START", "CALL.RECORD_STOP"):
            data = self.get_data_record()
        elif name == "CALL.END":
            data = self.get_data_end()

        push(name, self.ID, data, self._company)


# WMS-15219
def check_data_limits(data):
    db_number_limit = 64
    destination = data.get("destination")
    if destination and len(destination) > db_number_limit:
        logger.warning("Invalid destination number: {}".format(destination))
        data["destination"] = destination[:db_number_limit]
    callee = data.get("callee")
    if callee:
        phone = callee.get("phone")
        if phone and len(phone) > db_number_limit:
            logger.warning("Invalid callee phone: {}".format(phone))
            callee["phone"] = phone[:db_number_limit]


def push(name, id, data, company=None):
    ev = {
        "ext" if name.startswith("PRESENCE") else "id": id,
        "type": name,
        "time": int(get_time() * 1000),
        "pbx": Common.PBX_SERIAL if AppConfig.get("driver") == ConnectorType.CwConnector else Common.PBX_HOSTNAME
    }

    if not AppConfig.get("use_pcap_as_source"):
        ev["company"] = "vox_%s" % company if company else Aws.get_company()

    if data:
        check_data_limits(data)
        ev["data"] = data

    log_data = AppConfig.get_log_data_by_id(ev["id"]) if "id" in ev else None
    if log_data:
        if log_data["company"]:
            ev["company"] = log_data["company"]
        if log_data["serial"]:
            ev["pbx"] = log_data["serial"]
    payload = json.dumps(ev, cls=Common.PbxDataEncoder)
    if name.startswith("CALL"):
        logfile = log_data["out_file"] if log_data and "out_file" in log_data else AppConfig.get("out_file")
        if logfile:
            with open(logfile, "a") as log_file:
                log_file.write("{}\n".format(payload))
    if name == "PRESENCE.TELEPHONY":
        logger.info(payload)
    else:
        logger.notice(payload)
    event_type = "calls" if name.startswith("CALL") else "presence"
    stream = AppConfig.get("aws_stream_name")
    if stream and AppConfig.checkMode(event_type):
        logger.debug("Sending {} event to the Kinesis {} stream".format(name, stream))
        put_response = Aws.push_data(payload, stream)
    # print("Response: {}".format(put_response))
