import os
import time
import pickle
import re
import datetime
from threading import Thread
from dataEngine import Logger, AppConfig, Driver

logger = Logger.get()
ConferencesList = {}


class Recording:
    __slots__ = ("last_activity_time", "conf_id", "channel")

    def __init__(self, conf_id, channel):
        self.last_activity_time = None
        self.conf_id = conf_id
        self.channel = channel

    def start(self):
        exten=re.findall(r'^SIP/(.*)-.*$', self.channel)
        Driver.send_action("ConfbridgeStartRecord", Conference=self.conf_id, RecordFile="xhopcallrec,%s" % exten[0])
        logger.notice("confdbg: recording has been started")

    def stop(self):
        Driver.send_action("ConfbridgeStopRecord", Conference=self.conf_id)
        logger.notice("confdbg: recording has been stopped")

class Conference:
    __slots__ = ("conf_id", "active_channels", "record")

    def __init__(self, conf_id):
        self.conf_id = conf_id
        self.active_channels = []
        self.record = None

    def __str__(self):
        return "confdbg: conference object (conference id %s). Status: %s (%s). Recorded: %s." % \
               (self.conf_id, "active" if self.is_active() else "inactive", len(self.active_channels),
                "yes" if self.record else "no")

    def is_active(self):
        return True if self.active_channels else False

    def is_recorded(self):
        return self.record is not None

    def add_speaker(self, channel):
        logger.notice("confdbg: channel %s has started talking" % channel)
        if channel not in self.active_channels:
            self.active_channels.append(channel)
        else:
            logger.warning("confdbg: channel '%s' has been already added to active channels list!" % channel)

        if self.record is None:
            self.record = Recording(self.conf_id, channel)
            self.record.start()

    def remove_speaker(self, channel):
        logger.notice("confdbg: channel %s has stopped talking" % channel)
        if channel not in self.active_channels:
            logger.warning("confdbg: channel '%s' is absent in active channels list!" % channel)
            return

        self.active_channels.remove(channel)
        if not self.is_active():
            self.record.last_activity_time = datetime.datetime.now()

        if self.record is None:
            logger.warning("confdbg: channel '%s' stopped talking, but the recording hasn't been started yet!" % channel)
            return


def get(conf_id):
    if conf_id not in ConferencesList:
        ConferencesList[conf_id] = Conference(conf_id)
    return ConferencesList[conf_id]


def stop_inactive_recordings():
    delay = AppConfig.get("conf_seconds_before_stop_recording")
    for conf_id in ConferencesList:
        if ConferencesList[conf_id].is_recorded() and not ConferencesList[conf_id].is_active() and \
                datetime.datetime.now() - ConferencesList[conf_id].record.last_activity_time > \
                datetime.timedelta(seconds=delay):
            ConferencesList[conf_id].record.stop()
            ConferencesList[conf_id].record = None


def recordings_timer():
    while True:
        time.sleep(5)
        stop_inactive_recordings()


def main():
    stop_recording = Thread(target=recordings_timer)
    stop_recording.daemon = True
    stop_recording.start()


main()
