import re
import json
import base64
import threading
from urllib.parse import unquote
from dataEngine import Conversation, Channel, Event, AppConfig, Logger, Common
from dataEngine.cw import Ldap, Presence


QUEUE_COLLECT_TIME = 1.0
re._pattern_type = re.Pattern
logger = Logger.get()
print = logger.debug

queue_calls = {}
queue_agents_unknown = {}
user_agents = {}
queue_call_events = {}
queue_cancel_events = {}

parser_dst = re.compile("\W*/?(\+?\d+)")
parser_mobility = re.compile(r"^.+?/(.+?)&.+?/\1@mobility")


def process_call_timer(lid):
    agents = queue_call_events.get(lid)
    if agents is None:
        return
    data = [{"agent": k, "info": v} for k,v in agents.items()]
    if data:
        msgdata = {"agents": data}
        logger.debug("Queue agents: %s - %s" % (lid, msgdata))
        Event.push("CALL.AGENTS_START", lid, msgdata)
    queue_call_events.pop(lid)


def process_cancel_timer(lid):
    agents = queue_cancel_events.get(lid)
    if agents is None:
        return
    data = [{"agent": k, "reason": v} for k,v in agents.items()]
    if data:
        msgdata = {"agents": data}
        logger.debug("Queue cancel timer agents: %s - %s" % (lid, msgdata))
        Event.push("CALL.AGENTS_CANCEL", lid, msgdata)
    queue_cancel_events.pop(lid)


def add_agent_call_event(lid, agent, info):
    if lid not in queue_call_events:
        queue_call_events[lid] = {}
        t = threading.Timer(QUEUE_COLLECT_TIME, process_call_timer, (lid,))
        t.start()
    queue_call_events[lid][agent] = info


def process_agent_remove(lid, agent, reason):
    if lid not in queue_cancel_events:
        queue_cancel_events[lid] = {}
        t = threading.Timer(QUEUE_COLLECT_TIME, process_cancel_timer, (lid,))
        t.start()
    queue_cancel_events[lid][agent] = reason


def stop_queue_calling(lid, agents):
    if lid.startswith("wildixbox-"):
        lid = lid[10:]
    present = queue_cancel_events.get(lid)
    if present is not None:
        agents.update(present)
        queue_cancel_events.pop(lid)
    # remove delayed starts
    started = queue_call_events.get(lid)
    if started is not None:
        delayed = set(agents) & set(started)
        for agent in delayed:
            started.pop(agent, None)
            agents.pop(agent, None)
    data = [{"agent": k, "reason": v} for k,v in agents.items()]
    if data:
        msgdata = {"agents": data}
        logger.debug("Queue cancel agents: %s - %s" % (lid, msgdata))
        Event.push("CALL.AGENTS_CANCEL", lid, msgdata)

def process_queue_events(event):
    uid = get_param(event, "UniqueID")

    if event.name == "Join":
        queue_calls[uid] = {}

    elif event.name == "AgentCalled":
        if uid in queue_calls:
            agent_called = get_param(event, "AgentCalled")
            if agent_called not in queue_calls[uid]:
                callee_channel = Channel.get_by_name(get_param(event, "DestinationChannel"))
                if callee_channel:
                    queue_calls[uid][agent_called] = callee_channel
                    user_info = callee_channel.get_info()
                    call_lid = callee_channel.get("lid")
                    add_agent_call_event(call_lid, agent_called, user_info)
                else:
                    queue_agents_unknown[get_param(event, "DestinationChannel")] = agent_called
                    logger.info("Can't process AgentCalled event (%s) as channel %s hasn't been created yet."
                                " The channel will be added to the waiting list." %
                                (event, get_param(event, "DestinationChannel")))

    elif event.name == "AgentConnect":  # someone has answered, thus other agent received cancel
        if uid in queue_calls:
            agents = {}
            for agent in queue_calls[uid]:
                if get_param(event, "Member") != agent:
                    agents[agent] = "answered elsewhere"
                else:
                    agents[agent] = "answered here"
            stop_queue_calling(uid, agents)
            queue_calls.pop(uid)

    elif event.name == "QueueCallerAbandon":  # send missed call to called agents if a caller hangs up
        if uid in queue_calls:
            reason = 'caller left queue'
            if "answered_elsewhere_flag" in queue_calls[uid]:
                reason = 'answered elsewhere'
                queue_calls[uid].pop("answered_elsewhere_flag")
            agents = {agent:reason for agent in queue_calls[uid]}
            stop_queue_calling(uid, agents)
            queue_calls.pop(uid)

    elif event.name == "AgentRingNoAnswer":
        if uid in queue_calls and get_param(event, "Member") in queue_calls[uid]:
            agent = get_param(event, "Member")
            call_lid = queue_calls[uid][agent].get("lid")
            process_agent_remove(call_lid, agent, "agent timeout")
            queue_calls[uid].pop(agent)
    else:
        logger.warning("Unknown queue event: %s" % event)


def handle_cel_hangup(amiEvent, convObj, UniqueID):
    logger.info("Processing Hangup event")
    role = convObj.get_participant_role(UniqueID)
    if not role:
        return
    callerObj = convObj.get("caller")
    calleeObj = convObj.get("callee")
    extra = get_param(amiEvent, "Extra")
    splitExtra = extra.split(",")
    causeCode = splitExtra[0]
    dialstatus = splitExtra[2]
    hangupBy = None
    causeStr = "".join(splitExtra[3:])
    if splitExtra[1]:
        if callerObj and callerObj.get("chanName") == splitExtra[1]:
            hangupBy = "caller"
        elif calleeObj and calleeObj.get("chanName") == splitExtra[1]:
            hangupBy = "callee"
    if hangupBy is None and convObj.get("mergeWith") is not None:
        if convObj.get("transferType") is not None:
            hangupBy = role
            causeStr = "transfer"
    # set hangupBy for cases, when PBX ended the call
    if not hangupBy:
        logger.debug("hangupBy is empty :( [%s]" % extra)
        if AppConfig.isTest() and not AppConfig.get("compare_traces"):
            hangupBy = "Unknown"
        else:
            hangupBy = "callee"
    convObj.set("who",hangupBy)
    convObj.set("cause",causeCode)
    convObj.set("causeStr",causeStr)
    convObj.mark_remove_participant(role)


def handle_cel_transfer(amiEvent, amiEventName, UniqueID, convObj):
    logger.info("Processing %s event" % amiEventName)
    transferType = re.findall("^(\w+)TRANSFER", amiEventName)[0].lower()
    if transferType == "blind":
        extra = get_param(amiEvent, "Extra").split(",")
        destination = extra[0]
    elif transferType == "attended":
        destination = get_param(amiEvent, "Exten")
    convObj.set("destination", destination)
    destInfo = Ldap.getUserInfo(destination)
    # PARK_START
    if "park_orbit" == destInfo.get("userKind"):
        convObj.conv_split("park")
    # BLIND TRANSFER
    elif transferType == "blind":
        transferrer = convObj.get_participant_role(UniqueID)
        logger.debug("transferrer: {}".format(transferrer))
        logger.debug("convObj: {}".format(convObj))
        convObj.mark_remove_participant(transferrer)

        if transferrer:
            convObj.set(transferrer, None)
        else:
            logger.warning("Something went wrong. We don't have chan (%s) for blind transfer event "
                           "in our conversation (%s)" % (UniqueID, convObj))
            return
        convObj.set("transferType", transferType)
        convObj.set("causeStr", "transfer")
        convObj.set("who", transferrer)
        convObj.conv_split("transfer", destination)

    # ATTENDED TRANSFER
    elif transferType == "attended":
        def switching_data(conv1, conv2):
            conv1_caller = conv1.get("caller")
            conv2_caller = conv2.get("caller")
            conv1_callee = conv1.get("callee")
            conv2_callee = conv2.get("callee")
            if conv1_caller == conv2_caller:
                return {"caller": conv2_callee}, {"caller": conv1_callee}
            if conv1_caller == conv2_callee:
                return {"caller": conv2_caller}, {"callee": conv1_callee}
            elif conv1_callee == conv2_callee:
                return {"callee": conv2_caller}, {"callee": conv1_caller}
            elif conv1_callee == conv2_caller:
                return {"callee": conv2_callee}, {"caller": conv1_caller}
            logger.warning("Something went wrong, we didn't find the same participant for attendant transfer")
            return ()

        mergeWith = None
        peerName = get_param(amiEvent, "Peer")
        if get_param(amiEvent, "Extra"):
            mergeWith = get_param(amiEvent, "Extra")
        elif peerName:
            mergeWith = Channel.get_by_name(peerName).get("lid")
        oppositeConvObj = Conversation.get(mergeWith)

        if not oppositeConvObj:
            logger.error("Something went wrong, we don't have oppositeConvObj for mergeWith: ({}), "
                         "convObj: ({})".format(mergeWith, convObj))
            return

        if convObj == oppositeConvObj:
            logger.warning("!!!!!TODO!!!!! This case doesn't implemented yet (it's about attendant transfer "
                           "via feature code) conv: ({})".format(convObj))
            return

        data = switching_data(convObj, oppositeConvObj)
        for i in range(len(data)):
            who = next(iter(data[i]))
            chan = data[i].get(who)
            conv = convObj if i == 0 else oppositeConvObj
            ID = oppositeConvObj.get("ID") if i == 0 else convObj.get("ID")
            conv.set("mergeWith", ID)
            conv.set("transferType", transferType)
            conv.set("causeStr", "transfer")
            conv.set("who", who)
            conv.conv_split("transfer", chan.get("phone"))
            conv.link_channel(who, chan)


def handle_cel_chan_start(amiEvent, UniqueID, LinkedID, convObj):
    logger.debug("Processing channel creation")
    CallerIDnum = get_param(amiEvent, "CallerIDnum")
    CallerIDname = get_param(amiEvent, "CallerIDname")
    destination = get_param(amiEvent, "Exten")
    context = get_param(amiEvent, "Context")
    chanName = get_param(amiEvent, "Channel")
    chanType = Channel.ChanType.SIP if chanName.startswith("SIP/") else Channel.ChanType.LOCAL
    remoteApp, isFromWmsNet = Conversation.pop_wmsnet_channels(chanName)

    if chanType == Channel.ChanType.LOCAL and not CallerIDnum:
        CallerIDnum = destination

    userInfo = {}
    if "mobility" == context:
        userInfo = Ldap.getUserInfo(destination)
        destination = "mobility"
        if not CallerIDnum:
            CallerIDnum = userInfo.get("mobile")
        if not CallerIDname:
            CallerIDname = userInfo.get("commonName")
    elif CallerIDnum:
        userInfo = Ldap.getUserInfo(CallerIDnum)

    if not CallerIDname:
        CallerIDname = CallerIDnum

    if userInfo:
        logger.debug("The information about the channel's user from LDAP: %s" % userInfo)
    chanObj = Channel.Channel(
        uid=UniqueID,
        lid=LinkedID,
        userID=userInfo.get("userID"),
        email=userInfo.get("email"),
        userExtension=userInfo.get("userExtension"),
        groupID=userInfo.get("groupID"),
        groupName=userInfo.get("groupName"),
        phone=CallerIDnum,
        name=userInfo.get("commonName") if "commonName" in userInfo else CallerIDname,
        chanType=chanType,
        chanName=chanName,
        context=context,
        userDepartment=userInfo.get("department")
    )
    Channel.add(chanObj)

    if get_param(amiEvent, "Channel") in user_agents:
        logger.debug("User agent '%s' for the channel '%s' was got from the cache" %
                     (user_agents.get(get_param(amiEvent, "Channel")), get_param(amiEvent, "Channel")))
        chanObj.set("device", user_agents.pop(get_param(amiEvent, "Channel")))

    logger.debug("New channel has been created: %s" % chanObj)

    # here we process situations when AgentCalled event was generated before CHAN_START event
    if chanName in queue_agents_unknown:
        if LinkedID in queue_calls:
            agent = queue_agents_unknown[chanName]
            queue_calls[LinkedID][agent] = chanObj
            user_info = chanObj.get_info()
            call_lid = chanObj.get("lid")
            add_agent_call_event(call_lid, agent, user_info)
        queue_agents_unknown.pop(chanName)
    if not convObj:
        convObj = Conversation.Conversation(LinkedID, chanObj, destination)
        if context == "mobility":
            convObj.set_flags("mobility")
        Conversation.add(convObj)
        logger.debug("New conversation was created based on this channel")
        if remoteApp:
            convObj.setWmsNet(chanName, remoteApp, isFromWmsNet)
    else:
        if remoteApp:
            convObj.setWmsNet(chanName, remoteApp, isFromWmsNet)
        wmsNetCall = convObj.getWmsNet(chanName) or convObj.getWmsNet(convObj.get("caller"))
        wmsNetCallApp = wmsNetCall["app"] if wmsNetCall else None
        if wmsNetCallApp == "Queue" and wmsNetCall["isFromWmsNet"] == True:
            pass
        elif wmsNetCall and wmsNetCall["isFromWmsNet"] == False:
            convObj.skip_events_for_wmsnet_calls = True
        elif chanObj.get("chanType") != Channel.ChanType.LOCAL:
            if convObj.get("app") not in ("Queue", "Page"):  # don't send callee update using Queue or Page apps or wms network calls
                if not convObj.get("caller") or convObj.get("caller").isRemoved(): # case blind transfer
                    logger.info("Caller of existing conversation (%s) was updated with this channel. "
                                "Could be only for blind transfer case" % convObj.get("ID"))
                    convObj.link_channel("caller", chanObj)
                    return
                elif not convObj.get("callee") or convObj.get("callee").isRemoved():
                    logger.debug("Callee of existing conversation (%s) was updated with this channel" % convObj.get("ID"))
                    convObj.link_channel("callee", chanObj)
                    return
        logger.info("Skip updating conversation (%s) based on this channel (%s). The possible reasons are: "
                    "channel is Local; processing of queue/page applications; "
                    "conversation already has a callee / caller" % (convObj, chanObj))

    if Common.check_license("xhoppers"):
        from dataEngine.cw import XHoppers
        XHoppers.add_chan(CallerIDnum, chanName, destination)


def handle_bridge(amiEvent, convObj):
    logger.debug("Processing BRIDGE event: %s" % amiEvent.name)
    UniqueID1 = get_param(amiEvent, "UniqueID1")
    UniqueID2 = get_param(amiEvent, "UniqueID2")
    chanObj1 = Channel.get(UniqueID1)
    chanObj2 = Channel.get(UniqueID2)

    convObj.conversation_bridged = True
    if convObj.skip_events_for_wmsnet_calls:
        logger.info("Skip process event bridge due to we try to call to wms-network and all related event must be sent on another PBX")
        return

    # The logic is the following: if uniqueid 1 or 2 of bridge event equals caller's or callee's channel of
    # current conversation, we process it; otherwise - skip as it is service bridge that shouldn't be
    # indicated for customers
    bridged = False
    if convObj.get("callee") and convObj.get("callee").get("uid") in (UniqueID1, UniqueID2):
        bridged = True
        if UniqueID1 == convObj.get("callee").get("uid"):
            logger.debug("UniqueID1 of the event belongs to the following conversation: %s" % convObj)
            if not convObj.get("caller") or convObj.get("caller").get("uid") != chanObj2.uid:
                logger.debug("Linking the following channel as a caller to this conversation: %s" % chanObj2)
                convObj.link_channel("caller", chanObj2)
            else:
                logger.debug("Skip linking the channel as this channel is already linked to this conversation")
        elif UniqueID2 == convObj.get("callee").get("uid"):
            logger.debug("UniqueID2 of the event belongs to the following conversation: %s" % convObj)
            if not convObj.get("caller") or convObj.get("caller").get("uid") != chanObj1.uid:
                logger.debug("Linking the following channel as a caller to this conversation: %s" % chanObj1)
                convObj.link_channel("caller", chanObj1)
            else:
                logger.debug("Skip linking the channel as this channel is already linked to this conversation")

    if convObj.get("caller") and convObj.get("caller").get("uid") in (UniqueID1, UniqueID2):
        bridged = True
        if UniqueID1 == convObj.get("caller").get("uid"):
            logger.debug("UniqueID1 of the event belongs to the following conversation: %s" % convObj)
            if not convObj.get("callee") or convObj.get("callee").get("uid") != chanObj2.uid:
                logger.debug("Linking the following channel as a callee to this conversation: %s" % chanObj2)
                convObj.link_channel("callee", chanObj2)
            else:
                logger.debug("Skip linking the channel as this channel is already linked to this conversation")
        elif UniqueID2 == convObj.get("caller").get("uid"):
            logger.debug("UniqueID2 of the event belongs to the following conversation: %s" % convObj)
            if not convObj.get("callee") or convObj.get("callee").get("uid") != chanObj1.uid:
                logger.debug("Linking the following channel as a callee to this conversation: %s" % chanObj1)
                convObj.link_channel("callee", chanObj1)
            else:
                logger.debug("Skip linking the channel as this channel is already linked to this conversation")

    if bridged:
        if chanObj1.get("chanType") != Channel.ChanType.LOCAL and \
                chanObj2.get("chanType") != Channel.ChanType.LOCAL:
            convObj.set("status", "talking")
        else:
            logger.debug("Skip changing call status to 'talking' as specified channel/channels are Local")
    elif convObj.getWmsNet(chanObj1) and convObj.getWmsNet(chanObj1)["app"] == "Queue":
        logger.debug("Linking the following channel as a callee to this conversation: %s" % chanObj2)
        convObj.link_channel("callee", chanObj2)
        convObj.set("status", "talking")
    elif convObj.getWmsNet(chanObj2) and convObj.getWmsNet(chanObj2)["app"] == "Queue":
        logger.debug("Linking the following channel as a callee to this conversation: %s" % chanObj1)
        convObj.link_channel("callee", chanObj1)
        convObj.set("status", "talking")
    else:
        logger.debug("Skip changing call status to 'talking' as channels with specified unique ids are "
                     "not members of any conversation (%s)" % convObj)

    for role in ("caller", "callee"):
        if convObj.get(role):
            if convObj.get(role).split_record():
                convObj.record_stop(convObj.get(role))
                convObj.get(role).splitRecordDst = None


def event_listener(event, **kwargs):
    amiEvent = event
    if AppConfig.get("show_events"):
        if AppConfig.get("log_file") or AppConfig.isDaemon():
            logger.debug("AMI Event:\r\n%s", amiEvent)
        else:
            logger.debug("AMI Event:\r\n%s%s%s" % (Logger.COLORS.get("dark_grey"), amiEvent, Logger.COLORS.get("reset")))
    try:
        # PresenceEvent
        if amiEvent.name in ("PresenceEvent", "PresenceTelephonyEvent", "PeerEntry"):
            if amiEvent.name == "PeerEntry" and get_param(amiEvent, "RealtimeDevice") != "yes":
                return
            logger.debug("Processing %s" % amiEvent.name)
            handle_presence(amiEvent)
            return

        if amiEvent.name in ("Join", "QueueCallerAbandon", "AgentCalled", "AgentRingNoAnswer", "AgentConnect"):
            logger.debug("Processing queue event: %s" % amiEvent.name)
            process_queue_events(amiEvent)
            return

        if amiEvent.name in ("ConfbridgeTalking", "ConfbridgeLeave") and AppConfig.get("conf_recording"):
            logger.debug("Processing confbridge event: %s" % amiEvent.name)
            process_confbridge_events(amiEvent)
            return

        if amiEvent.name == "SIPOriginateReply":
            if Common.check_license("xhoppers"):
                from dataEngine.cw import XHoppers
                XHoppers.process_event_originate_reply(
                    get_param(amiEvent, "Exten"),
                    get_param(amiEvent, "Call-ID"),
                    get_param(amiEvent, "User-Agent"),
                    get_param(amiEvent, "State")
                )
            return

        if amiEvent.name == "EventReg":
            logger.debug("Processing %s" % amiEvent.name)
            if Common.check_license("xhoppers"):
                from dataEngine.cw import XHoppers
                XHoppers.process_event_reg(
                    get_param(amiEvent, "Type"),
                    get_param(amiEvent, "Device"),
                    get_param(amiEvent, "UserAgent"),
                    get_param(amiEvent, "UaID")
                )
            return

        if amiEvent.name == "CoreShowChannel":
            logger.debug("Processing %s" % amiEvent.name)
            if Common.check_license("xhoppers"):
                from dataEngine.cw import XHoppers
                XHoppers.add_chan(
                    get_param(amiEvent, "CallerIDnum"),
                    get_param(amiEvent, "Channel"),
                    get_param(amiEvent, "Extension"),
                    get_param(amiEvent, "IsOnHold")
                )
            return

        if amiEvent.name in ("TelephoneHoldEvent", "TelephoneUnHoldEvent"):
            logger.debug("Processing %s" % amiEvent.name)
            if Common.check_license("xhoppers"):
                from dataEngine.cw import XHoppers
                XHoppers.toggle_hold_state(
                    get_param(amiEvent, "Exten"),
                    get_param(amiEvent, "Channel"),
                    "hold" if amiEvent.name == "TelephoneHoldEvent" else "unhold"
                )

        if amiEvent.name == "TelephoneChargerEvent":
            logger.debug("Processing %s" % amiEvent.name)
            if Common.check_license("xhoppers"):
                from dataEngine.cw import XHoppers
                XHoppers.set_charger_state(
                    get_param(amiEvent, "Exten"),
                    get_param(amiEvent, "User-agent"),
                    get_param(amiEvent, "State")
                )
            return

        LinkedID = get_param(amiEvent, "LinkedID")
        convObj = Conversation.get(LinkedID)

        if amiEvent.name in ("FROM_WMSNET_CALL", "TO_WMSNET_CALL"):
            Conversation.add_wmsnet_channels(
                get_param(amiEvent, "Channel"),
                get_param(amiEvent, "RemoteApp"),
                True if amiEvent.name == "FROM_WMSNET_CALL" else False)
            return

        if amiEvent.name == "SIPUserAgent":
            logger.debug("Processing SIPUserAgent event")
            if not get_param(amiEvent, "UserAgent"):
                logger.warning("SIPUserAgent event doesn't contain a user-agent: %s" % amiEvent)
                return

            uid = get_param(amiEvent, "UniqueID")
            if convObj and uid:
                logger.debug("Triggering update of the channel '%s' to set a new user agent: '%s'" %
                             (get_param(amiEvent, "Channel"), get_param(amiEvent, "UserAgent")))
                role = convObj.get_participant_role(uid)
                if role:
                    convObj.set_participant_info(role, {'device': get_param(amiEvent, "UserAgent")})
                elif Channel.get(uid):  # a call to a queue (channel hasn't been linked to the conversation yet)
                    Channel.get(uid).set("device", get_param(amiEvent, "UserAgent"))
                else:
                    logger.warning("Can't find a channel to add the information got from SIPUserAgent event.")
            else:
                logger.debug("User agent '%s' will be stored for future usage" % get_param(amiEvent, "UserAgent"))
                user_agents[get_param(amiEvent, "Channel")] = get_param(amiEvent, "UserAgent")
            return

        if convObj is None and get_param(amiEvent, "EventName") not in ("CHAN_START", "CHAN_END"):
            chanName = get_param(amiEvent, "Channel")
            if chanName and chanName.startswith("ConfBridgeRecorder"):
                return
            if amiEvent.name in ("CEL", "BRIDGE_START", "BRIDGE_UPDATE", "MusicOnHold", "MonitorStart",
                                 "UniqueIDMasquerade"):
                logger.warning("We got an event %s, but no one conversation found for LinkedID: '%s', chanName: %s" %
                               (amiEvent.name, LinkedID, chanName))
            else:
                logger.debug("The event is not CHAN_START, whereas conversation object was not found. Let's skip it "
                             "chanName: %s" % chanName)
            return

        if amiEvent.name == "CEL":
            handle_cel(amiEvent)
        elif amiEvent.name in ("BRIDGE_START", "BRIDGE_UPDATE"):
            handle_bridge(amiEvent, convObj)
        # HOLD/UNHOLD
        elif amiEvent.name == "MusicOnHold":
            logger.info("Processing MusicOnHold event")
            state = get_param(amiEvent, "State")
            if convObj.skip_moh_event:
                logger.info("Skip processing event MusicOnHold due to it's an application Queue")
                if state == "Stop":
                    convObj.skip_moh_event = False
                return
            # when callweaver starts/finishes Page application, it starts moh for a moment for service needs;
            # this information is useless for customers
            if not convObj.get("page"):
                if state == "Start":
                    convObj.set("status","hold")
                elif state == "Stop":
                    convObj.set("status","talking")
        # RECORD_START
        elif amiEvent.name == "MonitorStart":
            logger.info("Starting recording")
            fileName = get_param(amiEvent, "FileName")
            if fileName:
                fileName = fileName.split('/')[-1]
            convObj.record_start(
                chanObj=Channel.get(get_param(amiEvent, "Uniqueid")),
                fileName=fileName,
                isSystemRecord=Common.str_to_bool(get_param(amiEvent, "IsSystemRecord"))
            )
        # RECORD_STOP
        elif amiEvent.name == "MonitorStop":
            logger.info("Finishing recording")
            convObj.record_stop(Channel.get(get_param(amiEvent, "Uniqueid")))
        # UniqueIDMasquerade
        elif amiEvent.name == "UniqueIDMasquerade":
            logger.debug("Processing UniqueIDMasquerade event")
            UniqueID1 = get_param(amiEvent, "UniqueID1")
            UniqueID2 = get_param(amiEvent, "UniqueID2")
            Channel.uid_masquerade(UniqueID1, UniqueID2, LinkedID)

    except Exception as e:
        logger.exception("Exception during event_listener processing")
        Channel.print_current_state()
        Conversation.print_current_state()


def handle_cel(amiEvent):
    amiEventName = get_param(amiEvent, "EventName")
    UniqueID = get_param(amiEvent, "UniqueID")
    LinkedID = get_param(amiEvent, "LinkedID")
    convObj = Conversation.get(LinkedID)

    # CHAN_START
    if amiEventName == "CHAN_START":
        handle_cel_chan_start(amiEvent, UniqueID, LinkedID, convObj)
    # TRUNK_CALL
    elif amiEventName == "TrunkCall":
        trunk = unquote(get_param(amiEvent, "Extra")).split(",")
        logger.info("Processing TrunkCall event: %s" % dict(zip(("direction", "name"), trunk)))
        convObj.set("trunk", dict(zip(("direction", "name"), trunk)))
    # LOOKUP
    elif amiEventName == "LookUp":
        logger.info("Processing lookup event")
        extra = get_param(amiEvent, "Extra").split(",")
        lookupInfo = dict(zip(("name", "company", "phone", "email"), extra))
        if extra[4] in ("inc", "out"):
            convObj.set_participant_info("caller" if extra[4] == "inc" else "callee", lookupInfo)
        else:
            logger.warning("Lookup event with absent direction has been skipped")
    # Number normalization during "Dial the trunk" if lookup through dialplan application has been done
    elif amiEventName == "NumberNormalization":
        logger.info("Processing NumberNormalization event")
        normnum = get_param(amiEvent, "Extra")
        convObj.set_participant_info("callee", {"phone": normnum})
    # ADD_TAG
    elif amiEventName == "Tag":
        extra = get_param(amiEvent, "Extra")
        logger.info("Processing Tag cel event: %s" % extra)
        tags = [t for t in extra.replace(" ", "").split(",") if t] # WMS-14536 remove empty ("") tags
        if tags:
            convObj.set("tags", tags)
    # App (UPDATE/SPLIT)
    elif amiEventName == "App":
        logger.info("Processing App cel event")
        chanName = get_param(amiEvent, "Channel")
        extra = get_param(amiEvent, "Extra").split(",")
        wmsAppName = convObj.getWmsNet(chanName)["app"] if convObj.getWmsNet(chanName) else None
        appName = extra[0]
        destination = parser_dst.findall(extra[1])
        if chanName.startswith("SIP/") and wmsAppName != "Queue":
            logger.debug("App cel event args: %s" % extra)
            if destination:
                convObj.set_app(appName, destination[0])
            if appName == "Record" and len(extra) > 1:
                chanObj = Channel.get(UniqueID)
                if extra[1] == "pbx":
                    args = ",".join(extra[i] if extra[i] else "" for i in (2, len(extra) - 1))
                    chanObj.set_record(isSystemRecord=True, sysRecordArgs=args)
                elif extra[1] in ("pause", "unpause"):
                    chanObj.set_record_pause_state(True if extra[1] == "pause" else False)
            if appName == "ConfBridge":
                convObj.set("status", "talking")
                for role in ("caller", "callee"):
                    if convObj.get(role):
                        if convObj.get(role).split_record():
                            convObj.record_stop(convObj.get(role))
                            convObj.get(role).splitRecordDst = None
        else:
            logger.info("Skip processing wms-network App cel event due to %s" % ("we have call from app: '{}' of wms-network".format(wmsAppName) if wmsAppName else "channel variable of event is not started with 'SIP/'"))
        if appName == "Dial" and parser_mobility.match(extra[1]):
            convObj.set_flags("mobility")

    # ANSWER
    elif amiEventName == "ANSWER":
        logger.info("Processing Answer")
        application = get_param(amiEvent, "Application")
        if application == "VoiceMail":
            convObj.set("status", "talking")
        elif not convObj.skip_events_for_wmsnet_calls and not convObj.conversation_bridged:
            if application in ("Playback", "Background", "Read", "AppQueue", "Queue", "Answer"):
                convObj.set("status", "hold")
    # CALL_GROUP
    elif amiEventName == "CallGroup":
        queue = get_param(amiEvent, "Extra").split(",")
        logger.info("Processing CallGroup event: %s" % dict(zip(("name", "id"), queue)))
        convObj.set("service", queue[0])
        convObj.set("queue", dict(zip(("name", "id"), queue)))
        convObj.set_app("Queue", queue[1])
    # PAGING
    elif amiEventName == "Page":
        page = get_param(amiEvent, "Extra").split(",")
        logger.info("Processing Page event: %s" % dict(zip(("name", "id"), page)))
        convObj.set("page", dict(zip(("id", "name"), page)))
        convObj.set_app("Page", page[0])
    # FORWARD
    elif amiEventName == "FORWARD":
        logger.info("Processing Forward event")
        destination = parser_dst.findall(get_param(amiEvent, "Extra"))
        logger.debug("forward destination: %s" % destination[0])
        convObj.set("destination", destination[0])
        convObj.conv_split("forward", destination[0])
    # TRANSFER | PARK_START
    elif amiEventName in ("BLINDTRANSFER", "ATTENDEDTRANSFER"):
        handle_cel_transfer(amiEvent, amiEventName, UniqueID, convObj)
    # PICKUP | PARK_END
    elif amiEventName == "PICKUP":
        logger.info("Processing Pickup event: %s" % amiEventName)
        destination = get_param(amiEvent, "Exten")
        destInfo = Ldap.getUserInfo(destination)
        convObj.set("destination", destination)
        # PARK END
        if "park_orbit" == destInfo.get("userKind"):
            convObj.conv_split("park_end")
        # PICKUP
        else:
            mergeWith = Channel.get_by_name(get_param(amiEvent, "Peer")).get("lid")
            convObj.set("mergeWith", mergeWith)
            targetConvObj = Conversation.get(mergeWith)
            # set hangup cause and hangup By for target conversation
            targetConvObj.set("causeStr", "pickup")
            targetConvObj.set("who", "caller")
            convObj.conv_split("pickup")
    elif amiEventName == "PickupViaBridge":
        logger.debug("Processing Pickup event: %s" % amiEventName)
        mergeWith = UniqueID
        targetChanName = get_param(amiEvent, "Extra").split(",")[0]
        targetChan = Channel.get_by_name(targetChanName)
        targetConvObj = Conversation.get(targetChan.get("lid"))
        targetConvObj.set("mergeWith", mergeWith)
        convObj.set("causeStr", "pickup")
        convObj.set("who", "caller")

        if targetConvObj.get("ID") in queue_calls:
            logger.debug("Fixed the reason for call.agent events")
            queue_calls[targetConvObj.get("ID")]["answered_elsewhere_flag"] = True

        targetConvObj.conv_split("pickup")
    # SERVICE
    elif amiEventName == "Service":
        service = get_param(amiEvent, "Extra")
        logger.info("Processing service event: %s" % service)
        convObj.set("service", service)
    # HANGUP
    elif amiEventName == "HANGUP":
        handle_cel_hangup(amiEvent, convObj, UniqueID)
        chanName = get_param(amiEvent, "Channel")
        # for hangup event from ConfBridgeRecorder we need to close the conversation,
        #   due to we don't have CHAN_END and LINKEDID_END in case conf record was finished by AMI command (WMS-16737)
        if chanName and chanName.startswith("ConfBridgeRecorder"):
            Channel.end(UniqueID)
            Conversation.end(LinkedID)
    # CHAN_END
    elif amiEventName == "CHAN_END":
        logger.debug("Channel %s has been finished" % UniqueID)
        if Common.check_license("xhoppers"):
            from dataEngine.cw import XHoppers
            XHoppers.remove_chan(get_param(amiEvent, "Channel"))
        convObj.record_stop(Channel.get(UniqueID))
        Channel.end(UniqueID)
    # LINKEDID_END
    elif amiEventName == "LINKEDID_END":
        logger.debug("Conversation %s has been finished" % LinkedID)
        Conversation.end(LinkedID)


def process_confbridge_events(amiEvent):
    from dataEngine.cw import Conference

    if amiEvent.name == "ConfbridgeTalking":
        confobj = Conference.get(get_param(amiEvent, "Conference"))
        if get_param(amiEvent, "TalkingStatus") == "on":
            confobj.add_speaker(get_param(amiEvent, "Channel"))
        else:
            confobj.remove_speaker(get_param(amiEvent, "Channel"))
    elif amiEvent.name == "ConfbridgeLeave":
        confobj = Conference.get(get_param(amiEvent, "Conference"))
        confobj.remove_speaker(get_param(amiEvent, "Channel"))


def handle_presence(amiEvent):
    def _serialize(o):
        res = {}
        if o.get("untilTime"):
            res["until"] = o.get("untilTime")
        if o.get("location"):
            tmp = json.JSONDecoder().decode(o.get("location"))
            if "" not in tmp:
                res["location"] = tmp
        if o.get("message"):
            res["message"] = o.get("message")
        return res

    def process_status_message(msg):
        res = {}
        if msg:
            try:
                msg = base64.b64decode(msg).decode("utf-8")
                res = json.JSONDecoder(object_hook=_serialize).decode(msg)
            except Exception as e:
                logger.warning("Something wrong with JSON (%s), Exception: %s" % (msg, e))
        return res
    data = {}
    ext = get_param(amiEvent, "User") or get_param(amiEvent, "ObjectName")
    userInfo = Ldap.getUserInfo(ext)
    if not userInfo:
        return
    department = userInfo.get("department")
    if department:
        data["department"] = department
    data["group"] = userInfo.get("groupID")
    eventname = "PRESENCE"
    # Try to refresh all peers statuses one by one
    if amiEvent.name == "PeerEntry":
        status = Presence.get_status(ext, get_param(amiEvent, "TelephonyState"), get_param(amiEvent, "RegState"))
        # Send status in case it was changed
        if status:
            data["status"] = status
            Event.push(eventname+".TELEPHONY", ext, data)
        if get_param(amiEvent, "Wildix-State").upper() in ("ONLINE", "DND", "AWAY"):
            data["status"] = get_param(amiEvent, "Wildix-State").upper()
            data.update(process_status_message(get_param(amiEvent, "Wildix-Custom")))
            Event.push(eventname, ext, data)
    else:
        # Status has been changed by user / call
        status = Presence.get_status(ext, get_param(amiEvent, "State"), get_param(amiEvent, "RegState"), get_param(amiEvent, "Message"))
        # Send status in case it was changed
        if status:
            data["status"] = status
            if amiEvent.name == "PresenceTelephonyEvent":
                eventname += ".TELEPHONY"
            else:
                data.update(process_status_message(get_param(amiEvent, "Message")))
            Event.push(eventname, ext, data)


def get_param(amiEvent, name):
    val = amiEvent.keys.get(name)
    if not val:  # function returns None when variable's value is an empty string; we'd better revise it
        correction = {"UniqueID": "Uniqueid",
                      "LinkedID": "Linkedid"}
        val = amiEvent.keys.get(correction.get(name))
    if val:
        return val.replace("wildixbox-", "")
