from logging import Logger

import time
import json
import os
import shutil
from scapy.all import rdpcap
from threading import Thread
from dataEngine import AppConfig, Logger, Event
from dataEngine.Driver import get_mock as mock, BasePcapReader


logger = Logger.get()


def _get_in_file(file):
    name = file.replace("/", "_")
    return os.path.join(AppConfig.get("tmp_dir"), "%s" % (name))


def _get_out_file(file):
    name = file.replace("/", "_")
    return os.path.join(AppConfig.get("tmp_dir"), "res_%s" % (name))


def _prepare_to_comparison(file, json_file):
    json_file = file.replace(".pcap", ".json")
    in_file = _get_in_file(json_file)
    out_file = _get_out_file(json_file)
    shutil.copyfile(json_file, in_file)
    try:
        os.remove(out_file)
    except IOError:
        pass

    AppConfig.set("out_file", out_file)
    return in_file, out_file


class Worker():
    def __init__(self, process_pcap, files, index):
        self.thread = Thread(target=process_pcap, args=(files, index,))
        self.isFinished = False
        self.files = files

    def start(self):
        self.thread.start()

    def close(self):
        self.isFinished = True
        if self.thread.is_alive():
            self.thread.join()


class PcapReader(BasePcapReader):
    def __init__(self):
        super(BasePcapReader, self).__init__()
        self.workers = dict()
        self.exit = False
        self.changed_cases = list()
        self.total_cases = 0
        self.log_files = dict()


    def _stopWorkers(self, index):
        if index in self.workers.keys():
            self.workers[index].isFinished = True


    def _set_log_file(self, name, data):
        self.log_files[name] = data


    def _get_log_file(self, name):
        return self.log_files[name] if name in self.log_files else None

    def _process_pcap(self, files, index):
        while not mock().get_logged_state():
            if self.exit:
                break
            time.sleep(0.5)

        tmp_dir = AppConfig.get("tmp_dir")
        for file in files:
            file = file.strip()
            if not os.path.isfile(file):
                logger.warning("File {} not found".format(file))
                continue
            if self.exit:
                break
            ldap_file = file.replace(".pcap", ".ldap")

            if not AppConfig.get("read_pcap_simultaneously"):
                AppConfig.set("have_ldap_storage", False)
                self._init_data(file, index)

            # prepare files to comparison
            json_file = file.replace(".pcap", ".json")
            if os.path.isfile(json_file):
                AppConfig.set("use_pcap_as_source", True)
            else:
                json_file = None
            if AppConfig.get("compare_traces"):
                if not json_file:
                    continue
                self.total_cases += 1

            if os.path.isfile(ldap_file):
                from dataEngine.cw import Ldap
                if Ldap.read(ldap_file):
                    AppConfig.set("have_ldap_storage", True)
            logger.notice("%sStart reading of PCAP file: %s (use ldap storage: %s)%s" % (Logger.COLORS.get("violet"), file, AppConfig.get("have_ldap_storage"), Logger.COLORS.get("reset")))
            scapy_cap = rdpcap(file)
            timestamp = 0
            for packet in scapy_cap:
                if self.exit:
                    break

                if AppConfig.get("rewrite_traces"):
                    Event.set_time(packet.time)
                    time.sleep(0.01)

                p = packet.sprintf("{IP:{TCP:%TCP.sport%}}")
                if not p.isdigit():
                    continue
                srcport = int(p)
                if srcport != 5039:
                    continue
                payload = packet.sprintf("{IP:{TCP:%Raw.load%}}").replace("'", "").replace("\\r", "\r").replace("\\n", "\n").replace("\\\\", "\\")
                if payload.startswith("Event"):
                    if AppConfig.get("realtime"):
                        if timestamp > 0:
                            sleep = packet.time - timestamp
                            time.sleep(sleep)
                    mock().send(payload)
                    timestamp = packet.time
            time.sleep(0.5)

        self._stopWorkers(index)


    def _compare_traces(self, pcap_file, index):
        tmp_dir = AppConfig.get("tmp_dir")
        json_file = pcap_file.replace(".pcap", ".json")

        log_file = self._get_log_file(pcap_file)
        in_file = log_file["in"]
        out_file = log_file["out"]
        try:
            with open(in_file, "r") as jf:
                orig_data = jf.readlines()
            with open(out_file, "r") as jf:
                new_data = jf.readlines()
        except Exception as e:
            logger.error(e)
            return
        orig_data_raw = ""
        new_data_raw = ""
        for i in range(0, max(len(orig_data), len(new_data))):
            if i < len(orig_data):
                orig_event = json.loads(orig_data[i])
                orig_event.pop("time")
                orig_data_raw = "%s%s\r\n" % (orig_data_raw, json.dumps(orig_event, sort_keys=True))

            if i < len(new_data):
                new_event = json.loads(new_data[i])
                new_event.pop("time")
                new_data_raw = "%s%s\r\n" % (new_data_raw, json.dumps(new_event, sort_keys=True))

        if orig_data_raw != new_data_raw:
            # logger.warning("The changed case is: %s" % pcap_file)
            with open(os.path.join(tmp_dir, "original"), "a") as original:
                original.write("%s\r\n" % pcap_file)
                original.write(orig_data_raw)
                original.write("\r\n")
            with open(os.path.join(tmp_dir, "final"), "a") as final:
                final.write("%s\r\n" % pcap_file)
                final.write(new_data_raw)
                final.write("\r\n")

            if AppConfig.get("rewrite_traces"):
                shutil.copyfile(out_file, json_file)

        os.remove(out_file)
        os.remove(in_file)

        if orig_data_raw != new_data_raw:
            return True


    def _init_data(self, file, index, workers=None):
        json_file = file.replace(".pcap", ".json")
        res = True
        try:
            with open(json_file, "r") as jf:
                data = jf.readlines()
                first_event = json.JSONDecoder().decode(data[0])
        except IOError:
            logger.error("Json file (%s) with the original results doesn't exist, nothing to compare." % json_file)
            return None
        except IndexError:
            logger.error("Json file (%s) with the original results is empty, nothing to compare." % json_file)
            return None
        except json.decoder.JSONDecodeError:
            logger.error("The first line of json file (%s) doesn't contain valid json." % json_file)
            return None
        except Exception as e:
            logger.error("Unknown problem during reading json file (%s): %s" % (json_file, str(e)))
            return None
        else:
            log_data = dict()
            log_data["serial"] = first_event.get("pbx", "")
            log_data["company"] = first_event.get("company", "")
            log_data["index"] = index
            if AppConfig.get("compare_traces"):
                in_file, out_file = _prepare_to_comparison(file, json_file)
                self._set_log_file(file, {"in": in_file, "out": out_file})
                log_data["out_file"] = out_file
            added = False
            for i in range(0, len(data)):
                id = json.JSONDecoder().decode(data[i]).get("id", None)
                if not id:
                    continue
                if AppConfig.get("read_pcap_simultaneously"):
                    if not AppConfig.get_log_data_by_id(id):
                        AppConfig.set_log_data_by_id(id, log_data)
                        workers[index] = [file]
                        added = True
                    elif not added:
                        index = AppConfig.get_log_data_by_id(id)["index"]
                        workers[index].append(file)
                        res = False
                        added = True
                else:
                    AppConfig.set_log_data_by_id(id, log_data)

        return res


    def close(self):
        self.exit = True
        for _, worker in self.workers.items():
            worker.close()


    def init(self, data):
        if not AppConfig.isTest():
            return

        files = dict()
        index = 0
        if AppConfig.get("read_pcap_simultaneously"):
            for file in data:
                if self._init_data(file, index, files):
                    index += 1
            for index, f in files.items():
                self.workers[index] = Worker(self._process_pcap, f, index)
        else:
            self.workers[index] = Worker(self._process_pcap, data, index)

        for _, worker in self.workers.items():
            worker.start()

        for _, worker in self.workers.items():
            while not worker.isFinished:
                time.sleep(1)

        if AppConfig.get("read_pcap_simultaneously"):
            time.sleep(5)

        # compare files
        if AppConfig.get("compare_traces"):
            for index, worker in self.workers.items():
                for f in worker.files:
                    if os.path.isfile(f):
                        if self._compare_traces(f, index):
                            self.changed_cases.append(f)
            logger.error("The number of changed cases is: %s (total: %s)" % (len(self.changed_cases), self.total_cases))
            if len(self.changed_cases) > 0:
                logger.error("Here is the list:\r\n%s\r\n" % "\r\n".join(self.changed_cases))
                logger.error("You can compare changed cases manually with the following command:\r\n\r\n"
                             "diff -u %s/original %s/final\r\n" % (AppConfig.get("tmp_dir"), AppConfig.get("tmp_dir")))

        if not AppConfig.isStopped():
            AppConfig.set("stop", True)