import httplib2
import json
import time
import os
from threading import Thread
import queue
import re
from dataEngine import AppConfig, Logger, Common, Driver

logger = Logger.get()
USER_AGENT = os.environ.get("USER_AGENT", "Wildix W-AIR-SP/HS,Wildix W-AIR-SBR2/HS")
CONF_EXTEN = os.environ.get("CONF_EXTEN", "*Wairhotline*")
UsersList = {}
ChannelsList = {}
RETRY_TIMEOUT = 3.0
queue_send_action = None
send_action_thread = None
periodic_checker = None

message_regex = re.compile(r".*Call-ID: (.*)\r")

class User(Common.PbxDataObject):
    __slots__ = ("user_id", "channels", "foreign_channels", "contact", "charger_state", "call_attempts", "hold", "call_id", "originate_state")

    def __init__(self, user_id):
        self.user_id = user_id
        self.channels = []
        self.foreign_channels = []
        self.contact = None
        self.charger_state = None
        self.call_attempts = 0
        self.hold = False
        self.call_id = None
        self.originate_state = None

    def __str__(self):
        return "user id: {}\nchannels: {}\nforeign_channels: {}\ncontact: {}\ncharger_state: {}\nhold: {}\ncall_id: {}\noriginate_state: {}\ncall_attempts: {}".\
            format(self.user_id, self.channels, self.foreign_channels, self.contact, self.charger_state, self.hold, self.call_id, self.originate_state, self.call_attempts)

    def get(self, param):
        return getattr(self, param, None)

    def set(self, param, value):
        setattr(self, param, value)

    def add_channel(self, chan_name):
        if chan_name not in self.channels:
            self.channels.append(chan_name)

    def remove_channel(self, chan_name):
        self.channels.remove(chan_name)

    def add_foreign_channel(self, chan_name):
        if chan_name not in self.foreign_channels:
            self.foreign_channels.append(chan_name)

    def remove_foreign_channel(self, chan_name):
        if chan_name in self.foreign_channels:
            self.foreign_channels.remove(chan_name)


def _get_user(user_id):
    if user_id not in UsersList:
        UsersList[user_id] = User(user_id)
    return UsersList[user_id]


def _is_xhoppers_device(useragent):
    if not useragent:
        return False
    for pattern in USER_AGENT.split(","):
        logger.debug("is_xhoppers_device. Useragent [ {} ] match pattern [ {} ]".format(useragent, pattern))
        if pattern in useragent:
            logger.debug("is_xhoppers_device: True.")
            return True


def _http_client(url):
    if not url:
        return
    h = httplib2.Http(cache=None, timeout=0.5)
    try:
        resp, content = h.request(url, "GET")
    except Exception as e:
        logger.error(e)
        return []
    if not content:
        logger.notice("Can't get result for request to {}".format(url))
        return []
    try:
        response = json.loads(content.decode('utf-8'))
    except Exception as e:
        logger.error(e)
        return []
    if response.get("type") == "error":
        logger.notice("Can't get data: {}".format(response.get("reason")))
        return []
    return response.get("result")


def _send_action_worker():
    while Common.check_license("xhoppers"):
        if AppConfig.isStopped():
            break
        item = queue_send_action.get()
        if item.get("action") == "try_put_to_conference":
            _put_to_conference(item.get("user"), item.get("timeout") if item.get("timeout") else 0)

        elif item.get("action") == "try_unhold":
            _unhold(item.get("user"))

        queue_send_action.task_done()


def _put_to_conference(user, try_after=0):
    if user is None:
        return
    if user.get("contact") is None:
        logger.debug("Don't try to put user %s to the conference (we have no contact info)" % user.get("user_id"))
        return
    if user.get("charger_state") == "on-charger":
        logger.debug("Don't try to put user %s to the conference (the user device is on charge)" % user.get("user_id"))
        return

    if len(user.get("foreign_channels")) > 0:
        logger.debug("Don't try to put user %s to the conference (we have active channels)" % user.get("user_id"))
        return

    if len(user.get("channels")) > 0:
        logger.debug("Don't try to put user %s to the conference (we have active channels)" % user.get("user_id"))
        return

    if user.get("call_attempts") > 10:
        logger.debug("We have to try again to put %s to the conference, but current originate state: %s" % (user.get("user_id"), user.get("originate_state")))
        user.set("call_id", None)
        user.set("originate_state", None)

    if user.get("call_id") is not None:
        user.set("call_attempts", user.get("call_attempts") + 1)
        logger.debug("Don't try to put user %s to the conference (REFER has been already sent)" % user.get("user_id"))
        return

    if try_after > 0:
        time.sleep(try_after)
    logger.debug("Put user {}, to the x-hoppers conference".format(user.get("user_id")))
    try:
        res = Driver.send_action_unsafe('SendRefer', Dest=user.get("user_id"), To=CONF_EXTEN, UaID=user.get("contact"))
        if res and res.find("Success") > 0:
            user.set("call_attempts", 0)
            call_id = message_regex.findall(res)[0]
            user.set("call_id", call_id)
            user.set("originate_state", "init")
    except:
        try_after = try_after + 1 if try_after < 10 else try_after
        logger.debug("We have a problem with start an action, let's pospone originate for user %s for %dsec" % (user.get("user_id"), try_after))
        queue_send_action.put({"action": "try_put_to_conference", "user": user, "timeout": try_after})


def _hungup_all_calls(user):
    for chan in user.get("channels"):
        logger.debug("Send hangup for channel {}".format(chan))
        Driver.send_action("Hangup", Channel=chan)


def _unhold(user):
    if len(user.get("foreign_channels")) != 0:
        return
    if len(user.get("channels")) != 1:
        return
    if user.get("hold"):
        chan_name = user.get("channels")[0]
        logger.debug("Unhold x-hoppers channel {}".format(chan_name))
        Driver.send_action("SIPNotifyBroadsoft", Channel=chan_name, Event="talk")
        print_user(user)


def _check_user_id(user_id):
    if not user_id or "unknown" in user_id:
        return False
    return True


def process_event_originate_reply(user_id, call_id, user_agent, state):
    user = _get_user(user_id)
    if user is None:
        return
    if user.get("call_id") != call_id:
        logger.warning("Something goes wrong, call-id (%s != %s)" % (call_id, user.get("call_id")))
        return
    if state.upper() == "200 OK":
        user.set("call_id", None)
        user.set("call_attempts", 0)
        user.set("originate_state", "completed")
    else:
        if state.startswith("503"):
            logger.warning("We have a problem with device %s, exten: %s" % (user_agent, user_id))
        user.set("originate_state", state)
    print_user(user)


def process_event_reg(event_type, user_id, useragent, contact):
    if not event_type or not _check_user_id(user_id) or not useragent\
            or not contact or not _is_xhoppers_device(useragent):
        return
    user = _get_user(user_id)
    if event_type == "add":
        logger.debug("User {} registered, contact {}".format(user_id, contact))
        user.set("contact", contact)
    elif event_type == "remove":
        logger.debug("User {} un-registered".format(user_id))
        user.set("contact", None)
        # ToDo need to decide, if we want to auto set "on-charger" after losing registration
        #user.set("charger_state", "on-charger")
        _hungup_all_calls(user)
    print_user(user)


def clean_all():
    for _, user in UsersList.items():
        for chan in user.get("channels"):
            remove_chan(chan)
        for chan in user.get("foreign_channels"):
            remove_chan(chan)
        user.set("call_id", None)
    init()


def add_chan(user_id, chan_name, exten, is_on_hold="false"):
    if not _check_user_id(user_id) or not chan_name:
        return
    user = None
    if exten == CONF_EXTEN:
        logger.debug("x-hoppers channel {} has been added".format(chan_name))
        user = _get_user(user_id)
        user.add_channel(chan_name)
        ChannelsList[chan_name] = user
    elif exten in UsersList or user_id in UsersList:
        logger.debug("x-hoppers foreign channel {} has been added".format(chan_name))
        user = _get_user(exten if exten in UsersList else user_id)
        user.add_foreign_channel(chan_name)
        ChannelsList[chan_name] = user
        return
    else:
        return
    user.set("charger_state", "off-charger")
    user.set("hold", True if is_on_hold == "true" else False)
    if user.get("hold"):
        queue_send_action.put({"action":"try_unhold", "user": user})

    if len(user.get("channels")) > 1:
        chan_name = user.get("channels")[-1:][0]
        logger.debug("Detected more than one channels in the x-hoppers conference for the user {}."
                     "Sending hangup for the channel {}".format(user_id, chan_name))
        Driver.send_action("Hangup", Channel=chan_name)
    print_user(user)


def remove_chan(chan_name):
    if chan_name not in ChannelsList:
        return
    user = ChannelsList.pop(chan_name, None)
    if user is None:
        return
    if chan_name in user.get("channels"):
        logger.debug("X-hoppers channel {} has been removed".format(chan_name))
        user.remove_channel(chan_name)
    if chan_name in user.get("foreign_channels"):
        logger.debug("X-hoppers foreign channel {} has been removed".format(chan_name))
        user.remove_foreign_channel(chan_name)
    queue_send_action.put({"action":"try_unhold", "user": user})
    queue_send_action.put({"action": "try_put_to_conference", "user": user, "timeout": RETRY_TIMEOUT})
    print_user(user)


def set_charger_state(user_id, useragent, charger_state):
    if not _check_user_id(user_id) or not useragent or not charger_state\
            or not _is_xhoppers_device(useragent):
        return
    logger.debug("X-hoppers user {}, state is {}".format(user_id, charger_state))
    user = _get_user(user_id)
    user.set("charger_state", charger_state)
    if charger_state == "on-charger":
        _hungup_all_calls(user)
    print_user(user)


def toggle_hold_state(user_id, channel, state):
    if not user_id or not channel or not state:
        return
    if user_id not in UsersList:
        return
    user = _get_user(user_id)
    if channel not in user.get("channels"):
        return
    user.set("hold", True if state == "hold" else False)
    if user.get("hold"):
        queue_send_action.put({"action":"try_unhold", "user": user})


def periodic_check():
    while Common.check_license("xhoppers"):
        time.sleep(5)
        for _, user in UsersList.items():
            queue_send_action.put({"action": "try_put_to_conference", "user": user})
            queue_send_action.put({"action":"try_unhold", "user": user})


def init():
    global periodic_checker, UsersList, queue_send_action, send_action_thread
    if not Common.check_license("xhoppers"):
        if periodic_checker and periodic_checker.is_alive():
            logger.debug("Init. XHoppers license is not active, stopping periodic checker thread")
            periodic_checker.join()
            periodic_checker = None
        for _, user in UsersList.items():
            user.set("contact", None)
            _hungup_all_calls(user)
        if len(UsersList) > 0:
            UsersList = {}

        if queue_send_action is not None:
            queue_send_action.join()
            queue_send_action = None

        if send_action_thread is not None and send_action_thread.is_alive():
            logger.debug("Init. XHoppers license is not active, stopping send action thread")
            send_action_thread.join()
            send_action_thread = None
        return

    if queue_send_action is None:
        queue_send_action = queue.Queue()
        send_action_thread = Thread(target=_send_action_worker, daemon=True)
        send_action_thread.start()

    logger.debug("Init. Get channels list")
    Driver.send_action("CoreShowChannels")
    logger.debug("Init. Get charger state")
    for item in _http_client("http://localhost:5060/v1/wair_charger_state/"):
        user_id = item["user"]
        if _check_user_id(user_id):
            _get_user(user_id).set("charger_state", item["state"])
    logger.debug("Init. Get registrations")
    for aor in _http_client("http://localhost:5060/v1/wair_devices/"):
        if _is_xhoppers_device("Wildix {}/{}".format(aor.get("base_model"), aor.get("model"))):
            user_id = aor.get("extension")
            if _check_user_id(user_id):
                _get_user(user_id).set("contact", aor.get("contact"))
    for _, user in UsersList.items():
        queue_send_action.put({"action": "try_put_to_conference", "user": user})

    if periodic_checker is None:
        logger.debug("Init. Starting Xhoppers periodic checker thread")
        periodic_checker = Thread(target=periodic_check, daemon=True, name="XHoppers periodic checker")
        periodic_checker.start()


def print_user(param):
    res = ""
    if param == "all":
        res = "\n\n".join(str(user) for _, user in UsersList.items())
    else:
        res = str(param)
    logger.debug(res)
    return res
