import json
import uuid
import base64
from dataEngine.fs import Kazoo
from dataEngine import Channel, Conversation, Logger, AppConfig

logger = Logger.get()


def generate_conversation_uuid():
    return str(uuid.uuid4())


def event_listener(event):
    body = event.headers
    event_name = body.get("Event-Name")
    if not event_name:
        logger.notice("Event name is missing, skipping the event..")
        return

    if AppConfig.get("show_events"):
        if AppConfig.get("log_file") or AppConfig.isDaemon():
            logger.debug("Event: %s\r\n%s" % (event_name, json.dumps(body, indent=4)))
        else:
            logger.debug("Event: %s\r\n%s%s%s" % (event_name, Logger.COLORS.get("dark_grey"),
                                                  json.dumps(body, indent=4), Logger.COLORS.get("reset")))

    account = body.get("variable_ecallmgr_Account-ID", "unknown")
    lid = body.get("Channel-Call-UUID", "")
    uid = body.get("Unique-ID", "")

    try:
        if account != "unknown":
            unknown_chan_obj = Channel.get(body.get("Channel-Call-UUID"), "unknown")
            if unknown_chan_obj:
                Channel.end(unknown_chan_obj.get("uid"), "unknown")
                Channel.add(unknown_chan_obj, account)
                user_info = Kazoo.get_user_info(body)
                if user_info:
                    unknown_chan_obj.set("userID", user_info["userID"])
                    unknown_chan_obj.set("email", user_info["email"])
                    unknown_chan_obj.set("userExtension", user_info["userExtension"])
                    unknown_chan_obj.set("phone", user_info["phone"])
                    unknown_chan_obj.set("name", user_info["name"])
                    logger.debug("Channel %s was updated with data got from Kazoo: %s" % (unknown_chan_obj.get("uid"),
                                                                                          user_info))

                conv_obj = Conversation.Conversation(generate_conversation_uuid(), unknown_chan_obj,
                                                     body.get("Caller-Destination-Number"),
                                                     Kazoo.get_account_name(account))
                Conversation.add(conv_obj, account)
                logger.debug("Channel %s was updated with a correct account (%s). A new "
                             "conversation has been created" % (unknown_chan_obj.get("uid"), account))

        if event_name == "CHANNEL_CREATE":
            logger.debug("Processing %s event. LID: %s. UID: %s. Channel-Name: %s" %
                         (event_name, body.get("Channel-Call-UUID", "None"), uid if uid else "None", body["Channel-Name"]))
            if body["Channel-Name"].startswith("loopback/"):
                logger.debug("Skip CHANNEL_CREATE event for loopback: %s (%s, %s)" % (body["Channel-Name"], uid, lid))
                return

            user_info = Kazoo.get_user_info(body)
            chan_obj = Channel.Channel(
                uid=uid,
                lid=lid,
                userID=user_info.get('userID', ""),
                email=user_info.get('email', ""),
                userExtension=user_info.get('userExtension', ""),
                groupID="",
                groupName="",
                phone=user_info.get('phone', body["Caller-Callee-ID-Number"] if body.get("Caller-Callee-ID-Number") else body["Caller-Caller-ID-Number"],),
                name=user_info.get('name', body["Caller-Callee-ID-Name"] if body.get("Caller-Callee-ID-Name") else body["Caller-Caller-ID-Name"]),
                chanType=1,
                chanName="%s-%s" % (body["Channel-Name"], uid),
                context="",
                userDepartment="",
            )
            Channel.add(chan_obj, account)
            logger.debug("New channel has been created: %s" % chan_obj)

            if account == "unknown":
                logger.debug("Skip creating of a new conversation as account is unknown")
                return

            if body.get("Caller-Destination-Number", "").startswith("*5"):
                logger.debug("Skip CHANNEL_CREATE event for Unpark feature code")
                return
            if body.get("Caller-Destination-Number", "") == "6683":
                logger.debug("Skip CHANNEL_CREATE event for Move feature code")
                return

            if body["Caller-Destination-Number"].find("intercept") >= 0:
                conv_obj = Conversation.Conversation(generate_conversation_uuid(), chan_obj,
                                                     body.get("Caller-Destination-Number"),
                                                     Kazoo.get_account_name(account))
                Conversation.add(conv_obj, account)
                target_conv_obj = Conversation.get_by_chan(body["Caller-Destination-Number"].rsplit(":")[-1], account)
                target_conv_obj.set("mergeWith", conv_obj.get("ID"))
                conv_obj.set("causeStr", "pickup")
                conv_obj.set("who", "caller")
                target_conv_obj.conv_split("pickup")
                Conversation.end(conv_obj.get("ID"), pbx=account)
                return

            conv_obj = Conversation.get_by_chan(uid, account)
            if not conv_obj:
                conv_obj = Conversation.get_by_chan(body.get("Other-Leg-Unique-ID"), account)

            if not conv_obj and body.get("Caller-Source", "") == "mod_loopback":
                if not body.get("variable_ecallmgr_Bridge-ID"):
                    logger.warning("CHANNEL_CREATE event for loopback channel doesn't contain "
                                   "'variable_ecallmgr_Bridge-ID':")
                    logger.warning(json.dumps(body, indent=4))
                conv_obj = Conversation.get_by_chan(body["variable_ecallmgr_Bridge-ID"], account)

            if not conv_obj:
                conv_obj = Conversation.Conversation(generate_conversation_uuid(), chan_obj,
                                                     body.get("Caller-Destination-Number"),
                                                     Kazoo.get_account_name(account))
                Conversation.add(conv_obj, account)
                logger.debug("A new conversation has been created")
            else:
                if conv_obj.get("causeStr") == "transfer" and conv_obj.get("transferType") not in ("blind", "attended"):
                    conv_obj.set("transferType", "blind")
                    conv_obj.conv_split("transfer")
                    conv_obj.transferType = None
                    conv_obj.causeStr = None

                if not conv_obj.get("caller") or conv_obj.get("caller").isRemoved():
                    logger.debug("Caller of the existing conversation (%s) was updated with this channel. "
                                 "It could be only for blind transfer case" % conv_obj.get("ID"))
                    conv_obj.link_channel("caller", chan_obj)

                elif not conv_obj.get("callee") or conv_obj.get("callee").isRemoved():
                    if conv_obj.get("app") == "Queue":
                        logger.debug("Skip linking a callee as it is a call to a ring group.")
                    else:
                        logger.debug("Callee of the existing conversation (%s) was updated with this channel."
                                     % conv_obj.get("ID"))
                        conv_obj.link_channel("callee", chan_obj)

                else:
                    logger.debug("Conversation for a new channel was found, but it already has both participants "
                                 "(looks like a user with several devices or pick up attempt)")

        elif event_name == "CHANNEL_EXECUTE":
            if event.headers["Application"] == "kz_multiset" \
                    and event.headers.get("Application-Data", "").find("voxloud_application") >= 0:
                conv_obj = Conversation.get_by_chan(uid, account)
                if not conv_obj:
                    logger.warning("Account: %s. voxloud_application variable has been set, but conversation doesn't "
                                   "exist: %s" % (account, event.headers["Application-Data"].split("=", 1)[1]))
                    return

                data = event.headers["Application-Data"].split("=", 1)[1].split(",")
                application = data[0]
                if application == "trunkcall":
                    logger.debug("Voxloud application -> trunkcall")
                    conv_obj.set("trunk", {"direction": data[1], "name": ""})
                elif application == "ringgroup":
                    logger.debug("Voxloud application -> ringgroup")
                    queue_exten = data[1]
                    queue_name = base64.b64decode(data[2]).decode("utf-8")
                    strategy = data[3]
                    if conv_obj.get("destination") == queue_exten and not conv_obj.get("queue"):
                        conv_obj.set("destination", "queue_%s" % queue_exten)
                    conv_obj.set("queue", {"name": queue_name, "id": queue_exten})
                    conv_obj.set_app("Queue", queue_exten)
                elif application == "voicemail":
                    logger.debug("Voxloud application -> voicemail")
                    conv_obj.set_app("VoiceMail", data[1])
                    conv_obj.set("status", "talking")
                elif application == "callforward":
                    logger.debug("Voxloud application -> callforward")
                    conv_obj.app = None
                    conv_obj.conv_split("forward")
                else:
                    logger.warning("Unknown application is set to 'voxloud_application' variable: %s" % application)

        elif event_name == "RECORD_START":
            logger.debug("Processing %s event. LID: %s. UID: %s" % (event_name, body.get("Channel-Call-UUID", "None"),
                                                                    body.get("Unique-ID", "None")))
            conv_obj = Conversation.get_by_chan(uid, account)
            if conv_obj:
                conv_obj.record_start(Channel.get(uid, account), body["Record-File-Path"])
                logger.info("Recording has been started")
            else:
                logger.info("The recording of a non-existent conversation was started "
                            "(it hasn't been implemented yet)")

        elif event_name == "RECORD_STOP":
            logger.debug("Processing %s event. LID: %s. UID: %s" % (event_name, body.get("Channel-Call-UUID", "None"),
                                                                    uid if uid else "None"))
            conv_obj = Conversation.get_by_chan(uid, account)
            if conv_obj:
                conv_obj.record_stop(Channel.get(uid, account), body["Record-File-Path"])
                logger.info("Recording has been stopped")
            else:
                logger.info("The recording of a non-existent conversation was stopped "
                            "(it hasn't been implemented yet)")

        elif event_name == "CHANNEL_BRIDGE":
            logger.debug("Processing %s event. LID: %s. UID: %s" % (event_name, body.get("Channel-Call-UUID", "None"),
                                                                    uid if uid else "None"))
            conv_obj = Conversation.get_by_chan(uid, account)
            if not conv_obj:
                conv_obj = Conversation.get_by_chan(body["Bridge-B-Unique-ID"], account)
                if not conv_obj:
                    logger.warning("Bridge attempt for a non-existent conversation. Account: %s" % account)
                    return

                logger.debug("It looks like pickup/unpark/move")
                conv_obj.link_channel("callee", Channel.get(body["Bridge-A-Unique-ID"], account))
                conv_obj.set("status", "talking")
                return

            if conv_obj.get("app") == "Queue" and not conv_obj.callee:
                logger.debug("It's a call from a ring group")
                conv_obj.link_channel("callee", Channel.get(body["Other-Leg-Unique-ID"], account))
                conv_obj.set("status", "talking")
                conv_obj.app = None
                return

            if not conv_obj.callee:
                logger.warning("Unknown case when a bridge is applied to a conversation without a callee, debugging "
                               "is needed. NOTE: the exception is calls made with autotests as they make external "
                               "calls avoiding Classound which is the impossible scenario for real customers. Account:"
                               " %s" % account)
                return

            if (conv_obj.caller.uid == body["Bridge-A-Unique-ID"] and conv_obj.callee.uid == body["Bridge-B-Unique-ID"]) \
                    or (conv_obj.callee.uid == body["Bridge-A-Unique-ID"] and conv_obj.caller.uid == body["Bridge-B-Unique-ID"]):
                if body.get("Answer-State") == "answered":
                    conv_obj.set("status", "talking")
            else:
                if not conv_obj.get("transferType"):
                    logger.debug("just a call (without a transfer)")
                    conv_obj.link_channel("callee", Channel.get(body["Other-Leg-Unique-ID"], account))
                    if body.get("Answer-State") == "answered":
                        conv_obj.set("status", "talking")
                else:
                    logger.debug("processing transfer")
                    if not conv_obj.get("caller") or conv_obj.get("caller").isRemoved():
                        conv_obj.link_channel("caller", Channel.get(body["Other-Leg-Unique-ID"], account))
                        conv_obj.set("status", "talking")
                    elif not conv_obj.get("callee") or conv_obj.get("callee").isRemoved():
                        conv_obj.link_channel("callee", Channel.get(body["Other-Leg-Unique-ID"], account))
                        conv_obj.set("status", "talking")
                    else:
                        logger.warning("Can't link a channel after a transfer attempt")

        elif event_name == "CHANNEL_ANSWER":
            logger.debug("Processing %s event. LID: %s. UID: %s" % (event_name, body.get("Channel-Call-UUID", "None"),
                                                                    uid if uid else "None"))
            conv_obj = Conversation.get_by_chan(uid, account)
            if not conv_obj:
                return

            if conv_obj.caller and conv_obj.callee and conv_obj.get("status") != "talking":
                conv_obj.set("status", "talking")

        elif event_name == "CHANNEL_DESTROY":
            logger.debug("Processing %s event. LID: %s. UID: %s" % (event_name, body.get("Channel-Call-UUID", "None"),
                                                                    uid if uid else "None"))
            Channel.end(uid, account)
            conv_obj = Conversation.get_by_chan(uid, account)
            if not conv_obj:
                conv_obj = Conversation.get_by_chan(body.get("Other-Leg-Unique-ID"), account)
            if not conv_obj:
                logger.info("Unknown channel has been destroyed")
                return

            conv_obj.mark_remove_participant(conv_obj.get_participant_role(uid))
            if (not conv_obj.get("caller") or conv_obj.get("caller").isRemoved()) \
                    and (not conv_obj.get("callee") or conv_obj.get("callee").isRemoved()):
                conv_obj.set("causeStr", body["Hangup-Cause"])
                conv_obj.set("who", "")
                logger.debug("Conversation '%s' has been finished", conv_obj.get("ID"))
                Conversation.end(conv_obj.get("ID"), pbx=account)

        elif event_name == "CHANNEL_HOLD":
            logger.debug("Processing %s event. LID: %s. UID: %s" % (event_name, body.get("Channel-Call-UUID", "None"),
                                                                    uid if uid else "None"))
            conv_obj = Conversation.get_by_chan(uid, account)
            if not conv_obj:
                logger.info("Hold attempt for a non-existent conversation")
                return
            conv_obj.set("status", "hold")

        elif event_name == "CHANNEL_UNHOLD":
            logger.debug("Processing %s event. LID: %s. UID: %s" % (event_name, body.get("Channel-Call-UUID", "None"),
                                                                    uid if uid else "None"))
            conv_obj = Conversation.get_by_chan(uid, account)
            if not conv_obj:
                logger.info("Unhold attempt for a non-existent conversation")
                return

            if conv_obj.get("status") != "connecting":
                conv_obj.set("status", "talking")

        elif event_name == "CUSTOM" and body["Event-Subclass"] in ("sofia::transferor", "sofia::replaced", "sofia::intercepted"):

            if body["Event-Subclass"] == "sofia::transferor":  # blind and attended transfers
                logger.debug("Processing %s event. LID: %s. UID: %s" % (body["Event-Subclass"],
                                                                       body.get("Channel-Call-UUID", "None"),
                                                                       uid if uid else "None"))
                if body.get("att_xfer_original_call_id"):
                    logger.debug("Skip 'sofia::transferor' as it is related to attendant transfer that was processed "
                                 "in 'sofia::replaced' event")
                    return

                conv_obj = Conversation.get_by_chan(uid, account)
                if not conv_obj:
                    logger.warning("Transfer attempt for a non-existent conversation")
                    return

                transferror = conv_obj.get_participant_role(uid)
                conv_obj.mark_remove_participant(transferror)
                conv_obj.set("causeStr", "transfer")
                conv_obj.set("who", transferror)
                conv_obj.status = "connecting"

            elif body["Event-Subclass"] == "sofia::replaced":  # attended transfer
                logger.debug("Processing %s event. LID: %s. UID: %s" % (body["Event-Subclass"],
                                                                       body.get("Channel-Call-UUID", "None"),
                                                                       uid if uid else "None"))
                conv1_obj = Conversation.get_by_chan(body["att_xfer_replaced_by"], account)
                conv2_obj = Conversation.get_by_chan(uid, account)

                conv1_obj.set("transferType", "attended")
                conv1_obj.set("mergeWith", conv2_obj.get("ID"))
                conv1_obj.conv_split("transfer")
                conv2_obj.set("transferType", "attended")
                conv2_obj.set("mergeWith", conv1_obj.get("ID"))
                conv2_obj.conv_split("transfer")

                if conv2_obj.caller.get("userId") == conv1_obj.caller.get("userId"):
                    conv1_obj.link_channel("caller", conv2_obj.callee)
                elif conv2_obj.caller.get("userId") == conv1_obj.callee.get("userId"):
                    conv1_obj.link_channel("callee", conv2_obj.callee)
                else:
                    logger.error("Can't determine the initiator of an attendant transfer. Account: %s. Conversation 1 "
                                 "id: %s. Conversation 2 id: %s" % (account,
                                                                    conv1_obj.get("ID") if conv1_obj else "None",
                                                                    conv2_obj.get("ID") if conv2_obj else "None"))
                    Channel.print_current_state()
                    Conversation.print_current_state()
                    logger.error(json.dumps(body, indent=4))

                conv2_obj.set("causeStr", "transfer")
                conv2_obj.set("who", "caller")  # might be changed to a some variable
                Conversation.end(conv2_obj.get("ID"), pbx=account)

                conv1_obj.transferType = None
                conv1_obj.causeStr = None

        else:
            logger.debug("The logic for processing %s event is absent, just skip it. "
                         "LID: %s. UID: %s" % (event_name, body.get("Channel-Call-UUID", "None"), uid if uid else "None"))
            if event_name == "CUSTOM":
                logger.debug("%s: %s. LID: %s. UID: %s" % (event.headers["Event-Name"], event.headers["Event-Subclass"],
                                                           body.get("Channel-Call-UUID", "None"),
                                                           uid if uid else "None"))
            elif event_name == "CHANNEL_EXECUTE":
                logger.debug("%s: %s (%s). LID: %s. UID: %s" % (event.headers["Event-Name"], event.headers["Application"],
                                                                event.headers.get("Application-Data", "None"),
                                                                body.get("Channel-Call-UUID", "None"),
                                                                uid if uid else "None"))
    except Exception as e:
        logger.exception("Exception during event_listener processing. Account: %s. Lid: %s. Uid: %s" % (account, lid,
                                                                                                        uid))
        Channel.print_current_state()
        Conversation.print_current_state()
