import os
import jwt
import json
import hashlib
import requests
from dataEngine import Logger
from datetime import datetime
from requests.exceptions import HTTPError, Timeout, ConnectTimeout, ConnectionError

logger = Logger.get()
kazoo_token = None
accounts = {}
kazoo_host = os.environ.get("KAZOO_API_HOST", "127.0.0.1")
kazoo_port = int(os.environ.get("KAZOO_API_PORT", "8000"))
kazoo_user = os.environ.get("KAZOO_API_USER", "")
kazoo_pass = os.environ.get("KAZOO_API_PASS", "")


def update_token():
    logger.debug("Trying to update Kazoo token")
    auth_str = "%s:%s" % (kazoo_user, kazoo_pass)
    cred = hashlib.md5(auth_str.encode('utf-8')).hexdigest()[:32]
    try:
        response = requests.put(url="http://%s:%s/v2/user_auth" % (kazoo_host, kazoo_port),
                                headers={"Content-Type": "application/json"},
                                timeout=3,
                                data=json.dumps({"data": {"credentials": cred, "account_name": "superadmin"}}))
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        logger.error("HTTP error during getting kazoo token: %s" % str(e))
        return False
    except requests.exceptions.RequestException as e:
        logger.error("Connection error during getting kazoo token: %s" % str(e))
        return False

    global kazoo_token
    try:
        kazoo_token = response.json()["auth_token"]
    except Exception as e:
        logger.error("Kazoo's response doesn't contain auth token: %s" % response.content)
        return False
    else:
        logger.notice("Kazoo token has been updated")
        return True


def request_kazoo(url):
    logger.debug("Trying to reach the following API endpoint: %s" % url)
    global kazoo_token
    if not kazoo_token:
        if not update_token():
            return
    else:
        try:
            token_expiration = jwt.decode(kazoo_token, verify=False).get("exp")
        except Exception as e:
            logger.exception("Can't get expiration time from JWT: %s" % str(e))
            return
        else:
            if token_expiration - datetime.now().timestamp() < 10:
                logger.info("Kazoo token has expired. It will be updated.")
                if not update_token():
                    return

    def do_request():
        try:
            response = requests.get(url=url, timeout=3, headers={"X-Auth-Token": kazoo_token})
        except (ConnectionError, ConnectTimeout, Timeout) as con_err:
            logger.error("Connection error during accessing Kazoo API (%s): %s" % (url, str(con_err)))
            return 408, False
        except HTTPError as http_err:
            logger.error("HTTP error during accessing Kazoo API (%s): %s" % (url, str(http_err)))
            return http_err.response.status_code, False
        except requests.exceptions.RequestException as e:
            logger.error("Unknown error during accessing Kazoo API: %s" % str(e))
            return 503, False
        else:
            return response.status_code, response

    status_code, response = do_request()
    if status_code == 401:
        update_token()
        status_code, response = do_request()

    if status_code in (200, 201) and response:
        try:
            return response.json()
        except Exception as e:
            logger.exception("ValueError during parsing Kazoo response: %s" % response.content)


def get_user_info(event):
    account = event.get("variable_ecallmgr_Account-ID")
    if not account:
        logger.debug("Creation of channel without account id set")
        return {}

    try:
        sip_peer_name = event["Channel-Name"].split("/")[2].split("@")[0]
    except Exception as e:
        logger.warning("Can't get peer name from the following Channel-Name: %s. Account: %s." % (event["Channel-Name"],
                                                                                                  account))
        return {}

    devices = request_kazoo("http://%s:%s/v2/accounts/%s/devices" % (kazoo_host, kazoo_port, account))
    if not devices:
        logger.warning("Kazoo API returns no devices for account %s" % account)
        return {}

    device = None
    for dev in devices.get("data"):
        if dev.get("username") == sip_peer_name:
            device = dev
            break

    if not device:
        logger.info("%s is not a user" % sip_peer_name)
        return {}

    if not device.get("owner_id"):
        logger.warning("Device [%s] doesn't have an owner. Account: %s" % (sip_peer_name, account))
        return {}

    data = request_kazoo("http://%s:%s/v2/accounts/%s/users/%s" % (kazoo_host, kazoo_port, account,
                                                                   device.get("owner_id", "")))
    if not data:
        return {}

    try:
        user_details = data["data"]
        user_data = dict()
        user_data['userID'] = device["owner_id"]
        user_data['email'] = user_details['email']
        user_data['userExtension'] = user_details['caller_id']['internal']['number']
        user_data['phone'] = user_details['caller_id']['internal']['number']
        user_data["name"] = "%s %s" % (user_details['first_name'], user_details['last_name'])
        return user_data
    except Exception:
        logger.error("Kazoo's response doesn't contain needed data. Account: %s. Device: %s. User id: %s. "
                     "Raw response: %s" % (account, sip_peer_name, device["owner_id"], data))
        return {}


def get_account_name(account_id):
    rv = ""
    global accounts
    if account_id in accounts:
        logger.debug("Domain for account with id [%s] is [%s] (got from cache)" % (account_id, accounts[account_id]))
        return accounts[account_id]

    data = request_kazoo("http://%s:%s/v2/accounts/%s" % (kazoo_host, kazoo_port, account_id))

    try:
        rv = data["data"]["realm"].split(".")[0]
    except Exception as e:
        logger.error("Error has happened during getting account name: %s. Kazoo response: %s" % (str(e), data))
    else:
        accounts[account_id] = rv
        logger.debug("Domain for account with id [%s] is [%s] (got from API)" % (account_id, accounts[account_id]))

    return rv
