#!/usr/bin/python3
import os
import sys
import time
import random
import argparse
import subprocess
from threading import Thread
from dataEngine import Aws, Conversation, Channel, AppConfig, Logger, Event, Common, HealthChecker
from dataEngine.Common import str_to_bool
from dataEngine.Driver import get_mock as mock, get_pcap_reader as pcap_reader, get_conn as conn, get_connector_type, init as DriverInit

logger = Logger.get()
print = logger.debug

# TODO
#1. add queue to push messages to kinesis, send it by batch to reduce payments

def stop(delay=None):
    msg = " with a delay of %d seconds" % delay if delay else ""
    logger.notice("Stop{}".format(msg))

    if AppConfig.isTest():
        pcap_reader().close()
        mock().close_connection()
    conn().close_connection()
    if not AppConfig.get("skip_socket"):
        from dataEngine import Socket
        Socket.close_connection()
    if AppConfig.get("dump"):
        from dataEngine import DumpWriter
        DumpWriter.close()

    if delay:
        time.sleep(delay)


def get_free_port():
    out = -1
    while out != 0:
        port = random.randint(7000, 65000)
        p1 = subprocess.Popen(["netstat", "-npa"], stdout=subprocess.PIPE)
        p2 = subprocess.Popen(["grep", str(port)], stdin=p1.stdout, stdout=subprocess.PIPE)
        p3 = subprocess.Popen(["wc", "-l"], stdin=p2.stdout, stdout=subprocess.PIPE)
        (out, err) = p3.communicate()
        out = int(out.decode("utf-8"))
    return port


def parse_args(argv):
    def str_to_loglevel(value):
        if value.isdigit():
            value = int(value)
            if value == 0:
                return "ERROR"
            if value == 1:
                return "WARNING"
            if value == 2:
                return "NOTICE"
            if value == 3:
                return "INFO"
            if value >= 4:
                return "DEBUG"
        elif value.upper() in ("ERROR", "WARNING", "INFO", "NOTICE", "DEBUG"):
            return value.upper()
        raise ValueError(f'{value} is not a valid logger value')

    def set_aws_stream_name(env):
        if env == "stage" or AppConfig.get("in_file"):
            AppConfig.set("aws_stream_name", Aws.AWS_STREAM_STAGE)
        elif env == "stable":
            AppConfig.set("aws_stream_name", Aws.AWS_STREAM_STABLE)
        elif env == "none":
            AppConfig.set("aws_stream_name", None)
        else:
            AppConfig.set("aws_stream_name", Aws.AWS_STREAM)


    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--in_file', default=None, nargs='+',
                        help='path to pcap input file collected by `tcpdump -s0 -ilo port $(netstat -npa | grep 5039 | grep python | cut -d: -f2 | cut -d' ' -f1) -w IN_FILE` or start script with parameter --dump to collect dump file. '
                             'It doesn\'t work with parameter --dump')
    parser.add_argument('-o', '--out_file', help='path to output file with json events', default=None)
    parser.add_argument('--driver', choices=('cw', 'fs'), default='cw', help='driver name to the call engine. Default value: \'cw\'')
    parser.add_argument('-rwo', '--rewrite_out_file', nargs='?', type=str_to_bool, const=True, default=None,
                        help='rewrite output file')
    parser.add_argument('-p', '--port', help='port of pseudo AMI interface', type=int, default=None)
    parser.add_argument('-d', '--daemon', help='indication flag about start script by systemd (it changes output from console to syslog)', type=str_to_bool, default=False, nargs='?', const=True)
    parser.add_argument('-l', '--loglevel', help='log level (0=ERROR | 1=WARNING | 2=NOTICE | 3=INFO | 4=DEBUG)', type=str_to_loglevel, default=None)
    parser.add_argument('-lf', '--log_file', help='path to log file (output to this file only; if --daemon option is not specified, output is available in console as well)', default=None)
    parser.add_argument('-cl', '--cache_ldap', help='cache ldap results data', type=str_to_bool, default=None, const=True, nargs='?')
    parser.add_argument('-fd', '--full_debug', help='enable full debug logging (include all external libraries)',
                        type=str_to_bool, default=False, const=True, nargs='?')
    parser.add_argument('-e', '--show_events', help='display all ami events (log level has to be set to debug)',
                        type=str_to_bool, default=None, const=True, nargs='?')
    parser.add_argument('-s', '--simultaneously', help='start read all pcap files simultaneously (doesn\'t work together with compare_traces mode)',
                        type=str_to_bool, default=False, const=True, nargs='?')
    parser.add_argument('-rt', '--realtime', help='use realtime mode with timeout for test mode (parameter --in_file)',
                        nargs='?', const=True, type=str_to_bool, default=False)
    parser.add_argument('--dump', nargs='?', help='Name of test case [default: `testcase`]. Write tcpdump of AMI communication (testcase.pcap), ldap responces (testcase.ldap) and events to json file (testcase.json). It doen\'t work with paramater --in_file. Must be started with root privilege',
                        default=None, const='testcase')
    parser.add_argument('-m', '--mode', nargs='*', choices=['auto', 'none', 'calls', 'presence'], default='auto',
                        help='working mode, controls what kind of events to be sent to the Kinesis (auto | none | calls | presence), multiple values allowed')
    parser.add_argument('-ct', '--compare_traces', help='It shows if generated events differ from original data. '
                                                        'It works with --in_file parameter only (original events '
                                                        'has to be stored in json file with the same name '
                                                        'in the same directory). "pde_tmp" directory will be created '
                                                        'in the directory you start script from.',
                        type=str_to_bool, default=None, const=True, nargs='?')
    parser.add_argument('-rwt', '--rewrite_traces', help='It works with --compare-traces parameter only. In this mode '
                                                        'traces that differ from original ones will be rewritten. '
                                                        'In this mode, the execution takes much more time due to '
                                                        'the necessity of generation plausible time codes.',
                        type=str_to_bool, default=None, const=True, nargs='?')
    parser.add_argument('--env', help='use specific AWS stream to send events', choices=('prod', 'stable', 'stage', 'none'), default='prod')
    parser.add_argument('-cr', '--conf_recording', help='experimental', nargs='?', const=True, type=str_to_bool,
                        default=False)
    parser.add_argument('-ss', '--skip_socket', help='Don\'t initialize UNIX socket', nargs='?', const=True, type=str_to_bool,
                        default=False)
    args = parser.parse_args()
    print("args: {}".format(args))

    AppConfig.set("cache_ldap", args.cache_ldap)

    mode = args.mode
    AppConfig.set("driver", get_connector_type(args.driver))
    AppConfig.set("skip_socket", args.skip_socket)

    if args.in_file:
        AppConfig.set("in_file", args.in_file)
        port = get_free_port()
        AppConfig.set("port", int(port))
        AppConfig.set("isTest", True)
        AppConfig.set("rewrite_out_file", True)
        AppConfig.set("realtime", args.realtime)
        if args.compare_traces:
            AppConfig.set("compare_traces", args.compare_traces)
            if args.rewrite_traces:
                AppConfig.set("rewrite_traces", args.rewrite_traces)
            tmp_dir = os.environ.get("TMP_DIR", "pde_tmp")
            AppConfig.set("tmp_dir", tmp_dir)
            import shutil
            if os.path.exists(tmp_dir):
                shutil.rmtree(tmp_dir)
            os.mkdir(tmp_dir)
    elif args.dump:
        AppConfig.set("rewrite_out_file", True)
        AppConfig.set("out_file", "{}.json".format(args.dump))
        AppConfig.set("dump", "{}.pcap".format(args.dump))
        AppConfig.set("ldap_file", "{}.ldap".format(args.dump))
        AppConfig.set("cache_ldap", True)
        mode = ['none']

    AppConfig.set("mode", mode)

    if args.loglevel:
        AppConfig.set("loglevel", args.loglevel)

    AppConfig.set("daemon", args.daemon)
    AppConfig.set("out_file", args.out_file)
    AppConfig.set("port", args.port)
    AppConfig.set("rewrite_out_file", args.rewrite_out_file)
    AppConfig.set("log_file", args.log_file)
    set_aws_stream_name(args.env)
    AppConfig.set("conf_recording", args.conf_recording)
    if args.simultaneously:
        AppConfig.set("read_pcap_simultaneously", args.simultaneously)
        if AppConfig.get("compare_traces"):
            logger.warning("simultaneously mode and compare_traces don't work at the same time")
            AppConfig.set("compare_traces", False)
    if args.show_events:
        AppConfig.set("show_events", args.show_events)
        AppConfig.set("loglevel", "DEBUG")
    if args.full_debug:
        AppConfig.set("full_debug", args.full_debug)
        AppConfig.set("loglevel", "DEBUG")


def cleanup_all_data():
    if Common.check_license("xhoppers"):
        from dataEngine.cw import XHoppers
        XHoppers.clean_all()
    Channel.end_all()
    Conversation.end_all()


def main(argv):
    parse_args(argv)
    Logger.init_logger()
    logger.info("Starting")

    if not AppConfig.get("skip_socket"):
        from dataEngine import Socket
        Socket.init_connection()
    if AppConfig.get("aws_stream_name"):
        Aws.init_client()

    if AppConfig.get("rewrite_out_file") and AppConfig.get("out_file") and os.path.isfile(AppConfig.get("out_file")):
        os.remove(AppConfig.get("out_file"))

    DriverInit(AppConfig.get("driver"))

    if AppConfig.isTest():
        mock().init_connection()

    logger.info("Current working mode is {}".format(AppConfig.get("mode")))
    if AppConfig.checkMode("none") and not AppConfig.isTest() and not AppConfig.get("dump"):
        logger.info("No connection to the engine interface is made and events are not sent to the Kinesis")
    else:
        HealthChecker.init()
        if not AppConfig.get("use_pcap_as_source"):
            call_live = Thread(target=Conversation.update_active_calls)
            call_live.daemon = True
            call_live.start()
        conn().init_connection(cleanup_all_data)
        if Common.check_license("xhoppers"):
            from dataEngine.cw import XHoppers
            XHoppers.init()
    if AppConfig.isTest():
        pcap_reader().init(AppConfig.get("in_file"))
    if AppConfig.get("dump"):
        from dataEngine import DumpWriter
        DumpWriter.init()

    try:
        while True:
            if AppConfig.isStopped():
                break
            time.sleep(5)
    except (KeyboardInterrupt, SystemExit):
        AppConfig.set("stop", True)

    stop()
    return AppConfig.get("exit_code")


if __name__ == "__main__":
    sys.exit(main(sys.argv))
